"""Functions for running naive and optimised encodings."""

from enum import Enum

import numpy as np
from numpy.typing import NDArray

from ferrmion.core import (
    anneal_enumerations,
    encode,
    encode_standard,
    standard_symplectic_matrix,
    topphatt_standard,
)

from ..hamiltonians import FermionHamiltonian, QubitHamiltonian


class CoreEncodingTypes(Enum):
    JordanWigner = "JW"
    BravyiKitaev = "BK"
    ParityEncoding = "PE"
    JKMN = "JKMN"


def jordan_wigner(fham: FermionHamiltonian) -> QubitHamiltonian:
    """Naive Jordan-Wigner Encoding.

    Args:
        fham (FermionHamiltonian): Hamiltonian to encode.

    Returns:
        QubitHamiltonian: Encoded Hamiltonian.
    """
    sigs, coeffs = fham.signatures_and_coefficients
    return encode_standard(
        encoding="JW",
        n_modes=fham.n_modes,
        n_qubits=fham.n_modes,
        signatures=sigs,
        coeffs=coeffs,
        constant_energy=fham.constant_energy,
    )


def bravyi_kitaev(fham: FermionHamiltonian) -> QubitHamiltonian:
    """Naive Bravyi-Kitaev Encoding.

    Args:
        fham (FermionHamiltonian): Hamiltonian to encode.

    Returns:
        QubitHamiltonian: Encoded Hamiltonian.
    """
    sigs, coeffs = fham.signatures_and_coefficients
    return encode_standard(
        encoding="BK",
        n_modes=fham.n_modes,
        n_qubits=fham.n_modes,
        signatures=sigs,
        coeffs=coeffs,
        constant_energy=fham.constant_energy,
    )


def parity(fham: FermionHamiltonian) -> QubitHamiltonian:
    """Naive Parity Encoding.

    Args:
        fham (FermionHamiltonian): Hamiltonian to encode.

    Returns:
        QubitHamiltonian: Encoded Hamiltonian.
    """
    sigs, coeffs = fham.signatures_and_coefficients
    return encode_standard(
        encoding="PE",
        n_modes=fham.n_modes,
        n_qubits=fham.n_modes,
        signatures=sigs,
        coeffs=coeffs,
        constant_energy=fham.constant_energy,
    )


def jkmn(fham: FermionHamiltonian) -> QubitHamiltonian:
    """Naive Jiang-Kalev-Mruczkiewicz-Neven Encoding.

    Args:
        fham (FermionHamiltonian): Hamiltonian to encode.

    Returns:
        QubitHamiltonian: Encoded Hamiltonian.
    """
    sigs, coeffs = fham.signatures_and_coefficients
    return encode_standard(
        encoding="JKMN",
        n_modes=fham.n_modes,
        n_qubits=fham.n_modes,
        signatures=sigs,
        coeffs=coeffs,
        constant_energy=fham.constant_energy,
    )


def jordan_wigner_topphatt(fham: FermionHamiltonian) -> QubitHamiltonian:
    """TOPP-HATT optimised Jordan-Wigner Encoding.

    Args:
        fham (FermionHamiltonian): Hamiltonian to encode.

    Returns:
        QubitHamiltonian: Encoded Hamiltonian.
    """
    return _standard_topphatt("JW", fham)


def bravyi_kitaev_topphatt(fham: FermionHamiltonian) -> QubitHamiltonian:
    """TOPP-HATT optimised Bravyi-Kitaev Encoding.

    Args:
        fham (FermionHamiltonian): Hamiltonian to encode.

    Returns:
        QubitHamiltonian: Encoded Hamiltonian.
    """
    return _standard_topphatt("BK", fham)


def parity_topphatt(fham: FermionHamiltonian) -> QubitHamiltonian:
    """TOPP-HATT optimised Parity Encoding.

    Args:
        fham (FermionHamiltonian): Hamiltonian to encode.

    Returns:
        QubitHamiltonian: Encoded Hamiltonian.
    """
    return _standard_topphatt("PE", fham)


def jkmn_topphatt(fham: FermionHamiltonian) -> QubitHamiltonian:
    """TOPP-HATT optimised Jiang-Kalev-Mruczkiewicz-Neven Encoding.

    Args:
        fham (FermionHamiltonian): Hamiltonian to encode.

    Returns:
        QubitHamiltonian: Encoded Hamiltonian.
    """
    return _standard_topphatt("JKMN", fham)


def _standard_topphatt(encoding: str, fham: FermionHamiltonian) -> QubitHamiltonian:
    sigs, coeffs = fham.signatures_and_coefficients
    ipow, sym = topphatt_standard(
        encoding=encoding,
        n_modes=fham.n_modes,
        n_qubits=fham.n_modes,
        signatures=sigs,
        coeffs=coeffs,
    )
    return encode(
        ipowers=ipow,
        symplectics=sym,
        signatures=sigs,
        coeffs=coeffs,
        constant_energy=fham.constant_energy,
    )


def jordan_wigner_annealed(
    fham: FermionHamiltonian,
    temperature: int | None = None,
    initial_guess: list[int] | None = None,
    coefficient_weighted: bool = True,
) -> QubitHamiltonian:
    """TOPP-HATT optimised Jordan-Wigner Encoding.

    Args:
        fham (FermionHamiltonian): Hamiltonian to encode.
        temperature (Optional[int]): Initial annealing temperature.
        initial_guess (Optional[list[int]]): Initial mode enumeration.
        coefficient_weighted (bool): True to minimise coefficient Pauli-weight.

    Returns:
        QubitHamiltonian: Encoded Hamiltonian.
    """
    return _standard_annealed(
        "JW", fham, temperature, initial_guess, coefficient_weighted
    )


def bravyi_kitaev_annealed(
    fham: FermionHamiltonian,
    temperature: int | None = None,
    initial_guess: list[int] | None = None,
    coefficient_weighted: bool = True,
) -> QubitHamiltonian:
    """TOPP-HATT optimised Bravyi-Kitaev Encoding.

    Args:
        fham (FermionHamiltonian): Hamiltonian to encode.
        temperature (Optional[int]): Initial annealing temperature.
        initial_guess (Optional[list[int]]): Initial mode enumeration.
        coefficient_weighted (bool): True to minimise coefficient Pauli-weight.

    Returns:
        QubitHamiltonian: Encoded Hamiltonian.
    """
    return _standard_annealed(
        "BK", fham, temperature, initial_guess, coefficient_weighted
    )


def parity_annealed(
    fham: FermionHamiltonian,
    temperature: int | None = None,
    initial_guess: list[int] | None = None,
    coefficient_weighted: bool = True,
) -> QubitHamiltonian:
    """TOPP-HATT optimised Parity Encoding.

    Args:
        fham (FermionHamiltonian): Hamiltonian to encode.
        temperature (Optional[int]): Initial annealing temperature.
        initial_guess (Optional[list[int]]): Initial mode enumeration.
        coefficient_weighted (bool): True to minimise coefficient Pauli-weight.

    Returns:
        QubitHamiltonian: Encoded Hamiltonian.
    """
    return _standard_annealed(
        "PE", fham, temperature, initial_guess, coefficient_weighted
    )


def jkmn_annealed(
    fham: FermionHamiltonian,
    temperature: int | None = None,
    initial_guess: list[int] | None = None,
    coefficient_weighted: bool = True,
) -> QubitHamiltonian:
    """TOPP-HATT optimised Jiang-Kalev-Mruczkiewicz-Neven Encoding.

    Args:
        fham (FermionHamiltonian): Hamiltonian to encode.
        temperature (Optional[int]): Initial annealing temperature.
        initial_guess (Optional[list[int]]): Initial mode enumeration.
        coefficient_weighted (bool): True to minimise coefficient Pauli-weight.

    Returns:
        QubitHamiltonian: Encoded Hamiltonian.
    """
    return _standard_annealed(
        "JKMN", fham, temperature, initial_guess, coefficient_weighted
    )


def _standard_annealed(
    encoding: str,
    fham: FermionHamiltonian,
    temperature: int | None,
    initial_guess: list[int] | None,
    coefficient_weighted: bool,
):
    sigs, coeffs = fham.signatures_and_coefficients
    ipow, sym = standard_symplectic_matrix(encoding, fham.n_modes)

    if temperature is None:
        temperature = fham.n_modes

    if isinstance(initial_guess, list):
        initial_guess: NDArray[np.uint] = np.array(initial_guess, dtype=np.uint)
    else:
        initial_guess: NDArray[np.uint] = np.array(
            [*range(fham.n_modes)], dtype=np.uint
        )

    anneal_enumerations(
        ipowers=ipow,
        symplectics=sym,
        signatures=sigs,
        coeffs=coeffs,
        temperature=temperature,
        initial_guess=initial_guess,
        coefficient_weighted=coefficient_weighted,
    )

    return encode(
        ipowers=ipow,
        symplectics=sym,
        signatures=sigs,
        coeffs=coeffs,
        constant_energy=fham.constant_energy,
    )



"""Init Optimize Submodule."""

from ..core import anneal_enumerations, topphatt_standard
from .bonsai import bonsai_algorithm
from .cost_functions import (
    coefficient_pauli_weight,
    distance_squared,
    minimise_mi_distance,
    pauli_weight,
)
from .enumeration.evolutionary import lambda_plus_mu
from .hatt import hamiltonian_adaptive_ternary_tree
from .huffman import huffman_ternary_tree
from .rett import reduced_entanglement_ternary_tree
from .topphatt import topphatt

__all__ = [
    "lambda_plus_mu",
    "minimise_mi_distance",
    "distance_squared",
    "coefficient_pauli_weight",
    "pauli_weight",
    "anneal_enumerations",
    "bonsai_algorithm",
    "huffman_ternary_tree",
    "reduced_entanglement_ternary_tree",
    "hamiltonian_adaptive_ternary_tree",
    "topphatt",
]



import numpy as np
import pytest
import scipy as sp
from ferrmion.encode.ternary_tree import (
    TernaryTree,
    TTNode,
    JW,
    JordanWigner,
    BK,
    BravyiKitaev,
    JKMN,
    ParityEncoding,
)
from ferrmion.utils import symplectic_hash, symplectic_unhash, symplectic_to_pauli
from openfermion import QubitOperator, get_sparse_operator
from openfermion.ops import InteractionOperator
from openfermion.transforms import jordan_wigner
from ferrmion.hamiltonians import molecular_hamiltonian
from ferrmion.core import standard_symplectic_matrix


@pytest.fixture
def six_mode_tree():
    return TernaryTree(n_modes=6, root_node=TTNode())


@pytest.fixture(scope="module")
def bonsai_paper_tree():
    tt = TernaryTree(n_modes=11)
    tt = tt.add_node("x")
    tt = tt.add_node("y")
    tt = tt.add_node("z")
    tt = tt.add_node("xx")
    tt = tt.add_node("xy")
    tt = tt.add_node("yx")
    tt = tt.add_node("yy")
    tt = tt.add_node("yz")
    tt = tt.add_node("zz")
    tt = tt.add_node("yzz")
    tt.enumeration_scheme = tt.default_enumeration_scheme()
    return tt


def test_standard_encoding_functions(six_mode_tree):
    # Test function aliases
    assert JW(6) == JordanWigner(6)
    assert BK(6) == BravyiKitaev(6)

    # Test TT function aliases
    assert six_mode_tree.JW() == JW(6)
    assert six_mode_tree.JordanWigner() == JordanWigner(6)
    assert six_mode_tree.BK() == BK(6)
    assert six_mode_tree.BravyiKitaev() == BravyiKitaev(6)
    assert six_mode_tree.JKMN() == JKMN(6)
    assert six_mode_tree.ParityEncoding() == ParityEncoding(6)

    # Test inequality by type
    assert JW(6) != BK(6)
    assert JW(6) != JKMN(6)
    assert JW(6) != ParityEncoding(6)
    assert BK(6) != JKMN(6)
    assert BK(6) != ParityEncoding(6)
    assert JKMN(6) != ParityEncoding(6)

    # Test inequality
    assert JW(6) != JW(5)
    assert JW(6) != JW
    assert JW(6) != "JW(6)"

    jw_different_enumeration = JW(6)
    jw_different_enumeration.enumeration_scheme["z"] = JW(6).enumeration_scheme["zz"]
    jw_different_enumeration.enumeration_scheme["zz"] = JW(6).enumeration_scheme["z"]
    assert JW(6) != jw_different_enumeration


def test_default_enumeration_scheme(six_mode_tree):
    assert six_mode_tree.default_enumeration_scheme() == {"": (0, 0)}
    jkmn = six_mode_tree.JKMN()
    assert jkmn.default_enumeration_scheme() == {
        "": (0, 0),
        "x": (1, 1),
        "y": (2, 2),
        "z": (3, 3),
        "xx": (4, 4),
        "xy": (5, 5),
    }


def test_invalid_enumeration_scheme(six_mode_tree):
    jkmn = six_mode_tree.JKMN()
    # Not enough qubit labels
    with pytest.raises(ValueError) as exc:
        jkmn.enumeration_scheme = {
            "": (0, 0),
            "x": (1, 1),
            "y": (2, 2),
            "z": (3, 3),
            "xx": (4, 4),
            "xy": (5, 4),
        }
    assert "Expected 6 qubit labels" in str(exc.value)

    # Not enough mode labels
    with pytest.raises(ValueError) as exc:
        jkmn.enumeration_scheme = {
            "": (0, 0),
            "x": (1, 1),
            "y": (2, 2),
            "z": (3, 3),
            "xx": (5, 4),
            "xy": (5, 5),
        }
    assert "Invalid mode labels" in str(exc.value)

    # Mode label not in range
    with pytest.raises(ValueError) as exc:
        jkmn.enumeration_scheme = {
            "": (6, 0),
            "x": (1, 1),
            "y": (2, 2),
            "z": (3, 3),
            "xx": (4, 4),
            "xy": (5, 5),
        }
    assert "Invalid mode labels" in str(exc.value)


def test_valid_enumeration_scheme(six_mode_tree):
    jkmn = six_mode_tree.JKMN()
    # We allow any qubit labels
    jkmn.enumeration_scheme = {
        "": (3, 10),
        "x": (2, 50),
        "y": (0, 30),
        "z": (1, 40),
        "xx": (4, 20),
        "xy": (5, 0),
    }


    jkmn.enumeration_scheme = {
        "": (3, 1),
        "x": (2, 5),
        "y": (0, 3),
        "z": (1, 4),
        "xx": (4, 2),
        "xy": (5, 0),
    }
    assert np.all(
        jkmn._build_symplectic_matrix()[1]
        == np.array(
            [
                [
                    False,
                    True,
                    False,
                    True,
                    False,
                    False,
                    False,
                    True,
                    False,
                    True,
                    False,
                    False,
                ],
                [
                    False,
                    True,
                    False,
                    True,
                    False,
                    False,
                    False,
                    True,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    False,
                    False,
                    False,
                    False,
                    True,
                    False,
                    False,
                    True,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    False,
                    False,
                    False,
                    False,
                    True,
                    False,
                    False,
                    True,
                    False,
                    False,
                    True,
                    False,
                ],
                [
                    False,
                    True,
                    False,
                    False,
                    False,
                    True,
                    False,
                    False,
                    True,
                    False,
                    False,
                    False,
                ],
                [
                    False,
                    True,
                    False,
                    False,
                    False,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    True,
                ],
                [
                    False,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    True,
                ],
                [
                    False,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    True,
                    False,
                    True,
                    False,
                    False,
                ],
                [
                    False,
                    True,
                    True,
                    False,
                    False,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    False,
                    True,
                    True,
                    False,
                    False,
                    True,
                    False,
                    False,
                    True,
                    False,
                    False,
                    False,
                ],
                [
                    True,
                    True,
                    False,
                    False,
                    False,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    True,
                ],
                [
                    True,
                    True,
                    False,
                    False,
                    False,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    True,
                ],
            ]
        )
    )


def test_bravyi_kitaev(six_mode_tree):
    tt = six_mode_tree.BK()
    assert tt.root_node.branch_strings == {
        "xxzy",
        "xxzx",
        "xxzz",
        "xzx",
        "xzy",
        "xzz",
        "xxy",
        "xxxx",
        "y",
        "xy",
        "z",
        "xxxz",
        "xxxy",
    }

    assert tt.root_node.child_strings == ["", "x", "xx", "xz", "xxx", "xxz"]

    assert tt.as_dict() == {
        "x": {
            "x": {
                "x": {"x": None, "y": None, "z": None},
                "y": None,
                "z": {"x": None, "y": None, "z": None},
            },
            "y": None,
            "z": {"x": None, "y": None, "z": None},
        },
        "y": None,
        "z": None,
    }

    assert tt.default_enumeration_scheme() == {
        "": (0, 0),
        "x": (1, 1),
        "xx": (2, 2),
        "xz": (3, 3),
        "xxx": (4, 4),
        "xxz": (5, 5),
    }

    assert tt.string_pairs == {
        "": ("xzz", "y"),
        "x": ("xxzz", "xy"),
        "xx": ("xxxz", "xxy"),
        "xz": ("xzx", "xzy"),
        "xxx": ("xxxx", "xxxy"),
        "xxz": ("xxzx", "xxzy"),
    }

    assert tt.branch_pauli_map == {
        "xxzy": "XXZIIY",
        "xxzx": "XXZIIX",
        "xxzz": "XXZIIZ",
        "xzx": "XZIXII",
        "xzy": "XZIYII",
        "xzz": "XZIZII",
        "xxy": "XXYIII",
        "xxxx": "XXXIXI",
        "y": "YIIIII",
        "xy": "XYIIII",
        "xxxz": "XXXIZI",
        "xxxy": "XXXIYI",
        "z": "ZIIIII",
    }

    assert tt.n_qubits == len(tt.root_node.child_strings)
    assert np.all(
        tt._build_symplectic_matrix()[1]
        == np.array(
            [
                [1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0],
                [1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1],
                [1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                [1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                [1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0],
                [1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0],
                [1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0],
                [1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                [1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0],
                [1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0],
                [1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1],
            ],
            dtype=np.int8,
        )
    )

    for line in tt._build_symplectic_matrix()[1]:
        assert np.all(line == symplectic_unhash(symplectic_hash(line), len(line)))


def tests_bonsai_paper_tree(bonsai_paper_tree):
    tt = bonsai_paper_tree
    assert tt.root_node.branch_strings == {
        "xyz",
        "zzy",
        "yyx",
        "yxz",
        "yzx",
        "yyy",
        "yzzx",
        "xyx",
        "xxx",
        "xxz",
        "yxx",
        "yzy",
        "xyy",
        "xxy",
        "yzzz",
        "yyz",
        "yxy",
        "zx",
        "zzz",
        "xz",
        "yzzy",
        "zzx",
        "zy",
    }

    assert tt.root_node.child_strings == [
        "",
        "x",
        "y",
        "z",
        "xx",
        "xy",
        "yx",
        "yy",
        "yz",
        "zz",
        "yzz",
    ]

    assert tt.as_dict() == {
        "x": {
            "x": {"x": None, "y": None, "z": None},
            "y": {"x": None, "y": None, "z": None},
            "z": None,
        },
        "y": {
            "x": {"x": None, "y": None, "z": None},
            "y": {"x": None, "y": None, "z": None},
            "z": {"x": None, "y": None, "z": {"x": None, "y": None, "z": None}},
        },
        "z": {"x": None, "y": None, "z": {"x": None, "y": None, "z": None}},
    }

    assert tt.default_enumeration_scheme() == {
        "": (0, 0),
        "x": (1, 1),
        "y": (2, 2),
        "z": (3, 3),
        "xx": (4, 4),
        "xy": (5, 5),
        "yx": (6, 6),
        "yy": (7, 7),
        "yz": (8, 8),
        "zz": (9, 9),
        "yzz": (10, 10),
    }

    assert tt.string_pairs == {
        "": ("xz", "yzzz"),
        "x": ("xxz", "xyz"),
        "y": ("yyz", "yxz"),
        "z": ("zx", "zy"),
        "xx": ("xxx", "xxy"),
        "xy": ("xyy", "xyx"),
        "yx": ("yxy", "yxx"),
        "yy": ("yyx", "yyy"),
        "yz": ("yzy", "yzx"),
        "zz": ("zzx", "zzy"),
        "yzz": ("yzzy", "yzzx"),
    }

    assert tt.branch_pauli_map == {
        "xyz": "XYIIIZIIIII",
        "zzy": "ZIIZIIIIIYI",
        "yyx": "YIYIIIIXIII",
        "yxz": "YIXIIIZIIII",
        "yzx": "YIZIIIIIXII",
        "yyy": "YIYIIIIYIII",
        "yzzx": "YIZIIIIIZIX",
        "xyx": "XYIIIXIIIII",
        "xxx": "XXIIXIIIIII",
        "xxz": "XXIIZIIIIII",
        "yxx": "YIXIIIXIIII",
        "yzy": "YIZIIIIIYII",
        "xyy": "XYIIIYIIIII",
        "xxy": "XXIIYIIIIII",
        "yzzz": "YIZIIIIIZIZ",
        "yyz": "YIYIIIIZIII",
        "yxy": "YIXIIIYIIII",
        "zx": "ZIIXIIIIIII",
        "xz": "XZIIIIIIIII",
        "yzzy": "YIZIIIIIZIY",
        "zzx": "ZIIZIIIIIXI",
        "zy": "ZIIYIIIIIII",
        "zzz": "ZIIZIIIIIZI",
    }

    assert tt.n_qubits == len(tt.root_node.child_strings)
    assert np.all(
        tt._build_symplectic_matrix()[1]
        == np.array(
            [
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1],
                [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1],
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0],
            ],
            dtype=np.int8,
        )
    )

    for line in tt._build_symplectic_matrix()[1]:
        assert np.all(line == symplectic_unhash(symplectic_hash(line), len(line)))

def test_default_mode_op_map(water_tt):
    assert np.all(water_tt.default_mode_op_map == [*range(water_tt.n_qubits)])

@pytest.mark.parametrize("n_modes", [1,5,10,20])
@pytest.mark.parametrize("encoding,name", [(JordanWigner, "JW"), (ParityEncoding, "PE"), (BravyiKitaev, "BK"), (JKMN, "JKMN")])
def test_core_standard_encodings(n_modes,encoding,name):
    n_modes = 20
    i,s = encoding(20)._build_symplectic_matrix()
    ci, cs = standard_symplectic_matrix(name,20)
    assert np.all(i==ci)
    assert np.all(s==cs)



"""Tests for functions in the optimize submodule."""

from ferrmion.optimize.hatt import hamiltonian_adaptive_ternary_tree, _qubit_term_weight

def test_reduce_hamiltonian():
    assert _qubit_term_weight([0], [0,1,2]) == 1
    assert _qubit_term_weight([1], [0,1,2]) == 1
    assert _qubit_term_weight([2], [0,1,2]) == 1

    assert _qubit_term_weight([0, 0], [0,1,2]) == 0
    assert _qubit_term_weight([0, 1], [0,1,2]) == 1
    assert _qubit_term_weight([0, 1, 1], [0,1,2]) == 1
    assert _qubit_term_weight([0, 1, 1, 1], [0,1,2]) == 1

    assert _qubit_term_weight([0, 1,2], [0,1,2]) == 0
    assert _qubit_term_weight([0, 1,2,2], [0,1,2]) == 1

    assert _qubit_term_weight([0, 0], [0,1,2]) == 0
    assert _qubit_term_weight([0, 0,0], [0,1,2]) == 1
    assert _qubit_term_weight([0, 0,0,0], [0,1,2]) == 0

    assert _qubit_term_weight([0, 3,5,6], [0,1,2]) == 1

def test_hatt():
    majorana_ham = {(0, 1): 0.5j, (2, 3): -0.5j, (4, 5): -0.5j, (2, 3, 4, 5): 0.5}
    n_modes = 3
    hatt = hamiltonian_adaptive_ternary_tree(majorana_ham, n_modes)
    assert hatt.as_dict() == {
        "x": {"x": 2, "y": 3, "z": 4},
        "y": 5,
        "z": {"x": 0, "y": 1, "z": 6},
    }
    assert hatt.enumeration_scheme == {"": (0, 2), "x": (1, 1), "z": (2, 0)}
    assert hatt.root_node.branch_majorana_map == {
        "y": 5,
        "xx": 2,
        "xy": 3,
        "xz": 4,
        "zx": 0,
        "zy": 1,
        "zz": 6,
    }



"""Tests for TOPP-HATT Algorithm."""

from ferrmion.optimize.topphatt import topphatt
from ferrmion.utils import fermionic_to_sparse_majorana
from ferrmion.encode.ternary_tree import (
    JordanWigner,
    BravyiKitaev,
    ParityEncoding,
    JKMN,
)
import numpy as np
import pytest
from ferrmion.core import topphatt_standard, encode
from ferrmion.optimize.huffman import huffman_ternary_tree
from ferrmion.optimize.hatt import hamiltonian_adaptive_ternary_tree, fast_hatt
from openfermion import QubitOperator, get_sparse_operator
from scipy.sparse.linalg import eigsh


@pytest.mark.parametrize("encoding", [JordanWigner, BravyiKitaev, ParityEncoding, JKMN])
def test_topphatt_preserves_topology(water_sparse_majorana, encoding):
    tree = encoding(14)
    tree = topphatt(water_sparse_majorana, tree)
    assert tree.root_node.child_strings == encoding(14).root_node.child_strings
    assert tree.root_node.branch_strings == encoding(14).root_node.branch_strings

def test_topphatt_huffman(water_sparse_majorana, water_data):
    test_tree = huffman_ternary_tree(water_data["ones"], water_data["twos"])
    initial_children = test_tree.root_node.child_strings
    initial_branches = test_tree.root_node.branch_strings
    topphatt_tree = topphatt(water_sparse_majorana, test_tree)
    assert topphatt_tree.root_node.child_strings == initial_children
    assert topphatt_tree.root_node.branch_strings == initial_branches

def test_topphatt_hatt(water_sparse_majorana, water_data):
    test_tree = hamiltonian_adaptive_ternary_tree(fermionic_to_sparse_majorana(((water_data["ones"],"+-"), (water_data["twos"], "++--"))), n_modes=14)
    initial_children = test_tree.root_node.child_strings
    initial_branches = test_tree.root_node.branch_strings
    topphatt_tree = topphatt(water_sparse_majorana, test_tree)
    assert topphatt_tree.root_node.child_strings == initial_children
    assert topphatt_tree.root_node.branch_strings == initial_branches


def test_topphatt_fasthatt(water_sparse_majorana, water_data):
    ones, twos = water_data["ones"], water_data["twos"]
    test_tree = fast_hatt(fermionic_to_sparse_majorana(((ones,"+-"), (twos, "++--"))), n_modes=14)
    initial_children = test_tree.root_node.child_strings
    initial_branches = test_tree.root_node.branch_strings
    topphatt_tree = topphatt(water_sparse_majorana, test_tree)
    assert topphatt_tree.root_node.child_strings == initial_children
    assert topphatt_tree.root_node.branch_strings == initial_branches


def test_topphatt_bonsai(water_sparse_majorana):
    pass


@pytest.mark.parametrize("encoding", ["JW", "BK", "PE", "JKMN"])
def test_core_topphatt_standard_h2_eigvals_equal_expected(encoding, h2_mol_data_sets):
    ones = h2_mol_data_sets["ones"]
    twos = h2_mol_data_sets["twos"]
    e_nuc = h2_mol_data_sets["constant_energy"]
    n_modes = ones.shape[0]
    ipow, sym = topphatt_standard(encoding, n_modes, n_modes, ["+-","++--"], [ones, twos])
    qham = encode(ipow, sym, ["+-","++--"],[ones, twos], e_nuc)


    ofop = QubitOperator()
    for k, v in qham.items():
        string = " ".join(
            [
                f"{char.upper()}{pos}" if char != "I" else ""
                for pos, char in enumerate(k)
            ]
        )
        ofop+= QubitOperator(term=string, coefficient=v)
    print(expected:=h2_mol_data_sets["eigvals"])
    diag, _ = eigsh(get_sparse_operator(ofop), k=2*n_modes, which="SA")
    print(diag)
    assert np.allclose(np.sort(diag), np.sort(h2_mol_data_sets["eigvals"]))



"""Base FermionQubitEncoding class."""
from inspect import signature

import logging
from abc import ABC, abstractmethod
from itertools import product

import numpy as np
from numpy.typing import NDArray

from ferrmion.core import hartree_fock_state, symplectic_product_map, encode, topphatt, anneal_enumerations
from ferrmion.utils import (
    icount_to_sign,
    pauli_to_symplectic,
    symplectic_product,
    symplectic_to_pauli,
    symplectic_to_sparse,
)
from ferrmion.hamiltonians import QubitHamiltonian, FermionHamiltonian

logger = logging.getLogger(__name__)


class FermionQubitEncoding(ABC):
    """Fermion Encodings for the Electronic Structure Hamiltonian in symplectic form.

    Attributes:
        one_e_coeffs (NDArray): One electron coefficients.
        two_e_coeffs (NDArray): Two electron coefficients.
        modes (set[int]): A set of modes.
        n_qubits (int): The number of qubits.

    Methods:
        default_mode_op_map: Get the default mode operator map.
        _build_symplectic_matrix: Build a symplectic matrix representing terms for each operator in the Hamiltonian.
        hartree_fock_state: Find the Hartree-Fock state of a majorana string encoding.
        _symplectic_to_pauli: Convert a symplectic matrix to a Pauli string.
        _pauli_to_symplectic: Convert a Pauli string to a symplectic matrix.
        fill_template: Fill a template with Hamiltonian coefficients.
        to_symplectic_hamiltonian: Output the hamiltonian in symplectic form.
        to_qubit_hamiltonian: Create qubit representation Hamiltonian.

    NOTE: A 'Y' pauli operator is mapped to -iXY so a (0+n)**3 term is needed.

    Example:
        >>> from ferrmion.encode.base import FermionQubitEncoding
        >>> class DummyEncoding(FermionQubitEncoding):
        ...     def _build_symplectic_matrix(self):
        ...         import numpy as np
        ...         return np.zeros(1), np.zeros((1, 2))
        >>> enc = DummyEncoding(2, 2)
        >>> enc.n_modes
        2
    """

    def __init__(
        self,
        n_modes: int,
        n_qubits: int,
    ):
        """Initialise encoding.

        Args:
            n_modes (int): Number of Fermion modes to encode.
            n_qubits (int): Number of Qubits used to encode.
            vacuum_state (NDArray | None): The vacuum state of the encoding.
        """
        self.n_modes = n_modes
        self.n_qubits = n_qubits
        self.default_mode_op_map = np.array([*range(self.n_modes)], dtype=np.uint)

    def __eq__(self, other: object) -> bool:
        """Checks if two encodings are exactly equivalent."""
        if isinstance(other, FermionQubitEncoding):
            if self.n_modes != other.n_modes:
                return False

            if self.n_qubits != other.n_qubits:
                return False

            left = self._build_symplectic_matrix()
            right = other._build_symplectic_matrix()

            if not np.all(left[0] == right[0]) or not np.all(left[1] == right[1]):
                return False

            return True
        else:
            return False

    @property
    def default_mode_op_map(self):
        """Create a default mode operator map for the tree."""
        return self._default_mode_op_map

    @default_mode_op_map.setter
    def default_mode_op_map(self, permutation: list[int]):
        """Set the default mode operator map.

        Args:
            permutation (list[int]): A list containing a permutation of mode indices.
        """
        logger.debug("Setting default mode operator map.")
        error_string = ""
        if set(permutation) != {*range(self.n_modes)}:
            error_string += "Default Mode op map does not cover all modes.\n"

        if error_string != "":
            logger.error(error_string)
            logger.error(permutation)
            raise ValueError(error_string)

        self._default_mode_op_map = np.array(permutation, dtype=np.uint)

    @property
    def vacuum_state(self):
        """Return the vacuum state."""
        return self._vacuum_state

    @vacuum_state.setter
    def vacuum_state(self, state: NDArray):
        """Validate and set the vacuum state.

        Args:
            state (NDArray): The vacuum state.
        """
        logger.debug("Setting vacuum state as %s", state)
        error_string = []
        state = np.array(state, dtype=np.float64)

        if len(state) != self.n_qubits:
            error_string.append("vacuum state must be length " + str(self.n_qubits))
        if state.ndim != 1:
            error_string.append("vacuum state must be vector (dimension==1)")

        if error_string != []:
            logger.error("\n".join(error_string))
            raise ValueError("\n".join(error_string))
        else:
            self._vacuum_state = state

    @abstractmethod
    def _build_symplectic_matrix(
        self,
    ) -> tuple[NDArray[np.uint8], NDArray[bool]]:
        """Build a symplectic matrix representing terms for each operator in the Hamitonian."""
        pass

    def encode_annealed(self, fham:FermionHamiltonian, temperature:int|None=None, initial_guess:list[int]|None=None, coefficient_weighted:bool=True):
        sigs, coeffs = fham.signatures_and_coefficients
        ipow, sym = self._build_symplectic_matrix()

        if temperature is None:
            temperature = fham.n_modes

        if isinstance(initial_guess, list):
            initial_guess: NDArray[np.uint] = np.array(initial_guess, dtype=np.uint)
        else:
            initial_guess: NDArray[np.uint] = np.array(
                [*range(fham.n_modes)], dtype=np.uint
            )

        anneal_enumerations(
            ipowers=ipow,
            symplectics=sym,
            signatures=sigs,
            coeffs=coeffs,
            temperature=temperature,
            initial_guess=initial_guess,
            coefficient_weighted=coefficient_weighted,
        )

        return encode(ipowers=ipow, symplectics=sym, signatures=sigs, coeffs=coeffs, constant_energy=fham.constant_energy)


    def encode(self, fham: FermionHamiltonian) -> QubitHamiltonian:
        logger.debug("Encoding fermionic Hamiltonian.")
        ipowers, symplectic = self._build_symplectic_matrix()
        signatures, coeffs = fham.signatures_and_coefficients
        return encode(
            ipowers=ipowers,
            symplectics=symplectic,
            signatures=signatures,
            coeffs=coeffs,
            constant_energy=fham.constant_energy
        )

    def hartree_fock_state(
        self,
        fermionic_hf_state: NDArray[bool],
        mode_op_map: list[int] | None = None,
    ):
        """Find the Hartree-Fock state of a majorana string encoding.

        This function calls to the rust implementatin in `src/lib.rs`.
        It assumes that the vacuum state is a single state vector, though the HF state may not be
        The global phase so that the first component state has 0 phase.

        Args:
            fermionic_hf_state (NDArray[int]): An array of mode occupations.
            mode_op_map (dict[int, int]): A dictionary mapping modes to sets of majorana strings i->(j,j+1).

        Returns:
            NDArray: The Hartree-Fock ground state in computational basis.
        """
        if mode_op_map is None:
            mode_op_map = self.default_mode_op_map

        return hartree_fock_state(
            self.vacuum_state,
            fermionic_hf_state,
            mode_op_map,
            self._build_symplectic_matrix()[0],
            self._build_symplectic_matrix()[1],
        )

    @staticmethod
    def _symplectic_to_pauli(
        symplectic: NDArray,
        ipower: int = 0,
    ) -> tuple[str, int]:
        """Convert a symplectic matrix to a Pauli string.

        Args:
            ipower (NDArray[np.uint]): power of i coefficient
            symplectic (NDArray): A symplectic vector.
        """
        return symplectic_to_pauli(symplectic, ipower)

    @staticmethod
    def _pauli_to_symplectic(
        pauli: str,
        ipower: int = 0,
    ) -> tuple[NDArray[bool], int]:
        """Convert a Pauli string to a symplectic matrix.

        Args:
            ipower (NDArray[np.uint]): power of i coefficient
            pauli (str): A Pauli-string.
        """
        return pauli_to_symplectic(pauli, ipower)

    @property
    def symplectic_product_map(self):
        """Calculate the product of symplectic terms and cache them."""
        logger.debug("Building symplectic product map")
        ipowers, symplectics = self._build_symplectic_matrix()
        return symplectic_product_map(ipowers, symplectics)

    def number_operator(
        self, mode: int
    ) -> list[tuple[str, NDArray, np.complexfloating]]:
        """Return the number operator of a mode for this encoding.

        Args:
            mode (int): The mode index to obtain a number operator for.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTee(4)
            >>> tree.number_operator(0)
        """
        return number_operator(self, mode)

    def edge_operator(
        self, edge_indices: tuple[int, int]
    ) -> list[tuple[str, NDArray, np.complexfloating]]:
        """Return the edge operator of a pair of modes for this encoding.

        Args:
            edge_indices (tuple[int, int]): The mode index to obtain a number operator for.

            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTee(4)
            >>> tree.edge_operator(0, 1)
        """
        return edge_operator(self, edge_indices)

    def majorana_product(
        self, product_indices: tuple[int, ...]
    ) -> tuple[np.ndarray[bool], np.complex64]:
        """Outputs the XZ-Matrix form of a product of majorana operators in an encoding.

        Args:
            product_indices (tuple[int,...]): A tuple of indices of majorana operators to apply.

        Return:
            np.ndarray[bool]: An XZ-encoded Pauli Operator.
            np.complex64: A complex float coefficient for the operator.
        """
        return majorana_product(self, product_indices)


def majorana_product(
    encoding: FermionQubitEncoding, product_indices: tuple[int, ...]
) -> tuple[np.ndarray[bool], np.complex64]:
    """Outputs the XZ-Matrix form of a product of majorana operators in an encoding.

    Args:
        encoding (FermionQubitEncoding): A valid encoding.
        product_indices (tuple[int,...]): A tuple of indices of majorana operators to apply.

    Return:
        np.ndarray[bool]: An XZ-encoded Pauli Operator.
        np.complex64: A complex float coefficient for the operator.

    Example:
    >>> jw = JordanWigner(4)
    >>> op, coeff = majorana_product(jw, (0,1,2,3))
    """
    ipowers, symplectics = encoding._build_symplectic_matrix()

    # all the majoranas have in imaginary factor which multiply
    # so they sum since we use powers
    ipower = sum(ipowers[m] for m in product_indices)

    # initial operator
    left = np.zeros(symplectics.shape[1], dtype=bool)

    logger.debug(f"{product_indices=},{ipower=}, {left=}")
    for m_index in product_indices:
        right = symplectics[m_index]
        iprod, left = symplectic_product(left, right)
        ipower += iprod
        logger.debug(f"{m_index=}, {ipower=}, {left=}")
    return left, icount_to_sign(ipower)


def number_operator(
    encoding: FermionQubitEncoding, mode: int
) -> list[tuple[str, NDArray, np.complexfloating]]:
    """Return the number operator for a given encoding and mode.

    Args:
        encoding (FermionQubitEncoding): A Fermion to qubit encoding object.
        mode (int): The mode index to obtain a number operator for.

    Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTee(4)
            >>> tree.number_operator(0)
    """
    return edge_operator(encoding, (mode, mode))


def edge_operator(
    encoding: FermionQubitEncoding, edge_indices: tuple[int, int]
) -> list[tuple[str, NDArray, np.complexfloating]]:
    """Return the edge operator for a given encoding and pair of modes.

    Args:
        encoding (FermionQubitEncoding): A Fermion to qubit encoding object.
        edge_indices (tuple[int, int]): The mode index to obtain a number operator for.

    Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTee(4)
            >>> tree.edge_operator(0,1)
    """
    return double_fermionic_operator(
        encoding=encoding, mode_indices=edge_indices, signature="+-"
    )


def double_fermionic_operator(
    encoding: FermionQubitEncoding, mode_indices: tuple[int, int], signature: str
) -> list[tuple[str, NDArray, np.complexfloating]]:
    """Returns the sparse pauli form of a double fermionic operator.

    Args:
        encoding (FermionQubitEncoding): A Fermion to qubit encoding object.
        mode_indices (tuple[int, int]): The mode indices to obtain a number operator for.
        signature (str): The fermionic operator signature, one of "++", "+-", "-+", "--".

    Returns:
        list[tuple[str, NDArray, np.complexfloating]]: A list of tuples each containing a Pauli string, its qubit indices and a complex coefficient.

    Example:
        >>> from ferrmion import TernaryTree
        >>> tree = TernaryTree(4)
        >>> tree.double_fermionic_operator((0,1), "+-")
        [('ZZ', array([0, 1]), 0.25+0j), ('YX', array([0, 1]), 0.25j), ('XY', array([0, 1]), -0.25j), ('II', array([0, 1]), 0.25+0j)]
    """
    match signature:
        case "++":
            signature_iterm = [1, -1j, -1j, -1]
        case "+-":
            signature_iterm = [1, 1j, -1j, 1]
        case "-+":
            signature_iterm = [1, -1j, 1j, 1]
        case "--":
            signature_iterm = [1, 1j, 1j, -1]
        case _:
            logger.error(
                "Operator signature can only contain + or -, %s not valid", signature
            )
            raise ValueError(
                "Operator signature can only contain + or -, %s not valid", signature
            )

    logger.debug("Finding double fermionic operator %s, %s", signature, mode_indices)
    if not set(mode_indices).issubset(set(range(encoding.n_modes))):
        logger.error("Edge operator indices invalid %s", mode_indices)
        raise ValueError("Edge operator indices invalid %s", mode_indices)

    icount, sym_products = encoding.symplectic_product_map
    m, n = mode_indices
    m = int(encoding.default_mode_op_map[m])
    n = int(encoding.default_mode_op_map[n])
    terms: list[tuple[str, NDArray, np.complex]] = [
        symplectic_to_sparse(
            sym_products[2 * m + l, 2 * n + r], icount[2 * m + l, 2 * n + r]
        )
        for l, r in product([0, 1], [0, 1])
    ]

    sparse_op = [
        (t[0], t[1], 0.25 * t[2] * si) for t, si in zip(terms, signature_iterm)
    ]
    logger.debug(f"Found operator {sparse_op}")
    return sparse_op



"""Contains the class for ternary tree nodes."""

import logging
from typing import Optional

import rustworkx as rx

logger = logging.getLogger(__name__)


class TTNode:
    """A node in a ternary tree.

    Attributes:
        parent (TTNode): The parent node.
        label (int | str): The qubit label.
        x (TTNode): The x child node.
        y (TTNode): The y child node.
        z (TTNode): The z child node.

    Methods:
        as_dict(): Convert the node to a dictionary.
        branch_strings(): Get the branch strings for the node.
        child_strings(): Get the child strings for the node.
        add_child(which_child, root_path): Add a child node to the current node.

    Simple Example:
        >>> from ferrmion.encode.ternary_tree_node import TTNode
        >>> node = TTNode()
        >>> node.add_child('x')
        >>> node.as_dict()
    """

    def __str__(self):
        """String representation of node."""
        return f"TTNode (root path:'{self.root_path}', Qubit Index:{self.qubit_label})"

    def __repr__(self):
        """String representation of node."""
        return f"TTNode (root path:'{self.root_path}', Qubit Index:{self.qubit_label})"

    def __init__(
        self,
        parent: Optional["TTNode"] = None,
        qubit_label: int | str | None = None,
    ):
        """Initialise a ternary tree node.

        Args:
            parent (TTNode | None): The parent node.
            root_path (str | None): The path from root to this node.
            qubit_label (int | str | None): The qubit label.
        """
        logger.debug(
            f"Creating TTNode with parent {parent} and qubit label {qubit_label}"
        )
        self.root_path = ""
        self.qubit_label = qubit_label
        self.parent: TTNode | None = parent
        self.x: TTNode | None = None
        self.y: TTNode | None = None
        self.z: TTNode | None = None
        self.z_ancestor: TTNode = self
        self.leaf_majorana_indices: dict[str, int | None] = {
            "x": None,
            "y": None,
            "z": None,
        }

    # def __str__(self) -> str:
    # return f"{self.as_dict()}"

    def as_dict(self) -> dict:
        """Return the node structure as a dictionary.

        Example:
            >>> from ferrmion.encode.ternary_tree_node import TTNode
            >>> node = TTNode()
            >>> node.add_child('x')
            >>> node.as_dict()
        """
        return as_dict(self)

    @property
    def branch_strings(self) -> set[str]:
        """Return a list of all branch strings for the node.

        Example:
            >>> from ferrmion.encode.ternary_tree_node import TTNode
            >>> node = TTNode()
            >>> node.add_child('x')
            >>> node.branch_strings
        """
        return set(branch_majorana_map(self).keys())

    @property
    def branch_majorana_map(self) -> dict[str, int]:
        """Create a map from branch strings to majorana indices.

        Returns:
            dict[str,int]: A map from branch strings to majorana operator indices.

        Example:
            >>> from ferrmion.encode.ternary_tree_node import TTNode, branch_strings
            >>> node = TTNode()
            >>> node.add_child('x')
            >>> branch_majorana_map(node)
        """
        return branch_majorana_map(root_node=self)

    @property
    def child_strings(self) -> list[str]:
        """Return a list of all child strings for the node.

        Example:
            >>> from ferrmion.encode.ternary_tree_node import TTNode
            >>> node = TTNode()
            >>> node.add_child('x')
            >>> node.child_strings
        """
        return sorted(child_strings(self, prefix=""), key=node_sorter)

    @property
    def child_qubit_labels(self) -> dict[str, int | None]:
        """Return a dict of sorted child nodes and their qubit label.

        Example:
            >>> from ferrmion.encode.ternary_tree_node import TTNode
            >>> node = TTNode()
            >>> node.add_child('x', qubit_label=5)
            >>> node.child_qubit_labels
        """
        return child_qubit_labels(self)

    def update_root_path(self, prefix: str) -> None:
        """Prefix the root path of a node and all its children.

        Args:
            prefix (str): String to prefix to root paths.
        """
        return update_root_path(root=self, prefix=prefix)

    @property
    def z_descendant(self) -> "TTNode":
        """Find the furthest z-descendant of a node."""
        return z_descendant(self)

    def add_child(
        self,
        which_child: str,
        child_node: Optional["TTNode"] = None,
        root_path: str | None = None,
        qubit_label: int | str | None = None,
    ) -> "TTNode":
        """Add a child node to the current node.

        Args:
            which_child (str): The child node to add.
            child_node (TTNode|None): A node object to set as the child.
            root_path (str): Path from root node.
            qubit_label (int | str): The qubit label.

        Returns:
            TTNode: The added child node

        Example:
            >>> from ferrmion.encode.ternary_tree_node import TTNode
            >>> node = TTNode()
            >>> node.add_child('x')
        """
        return add_child(
            self,
            which_child=which_child,
            child_node=child_node,
            qubit_label=qubit_label,
        )

    def to_rustworkx(self, with_leaves: bool = False) -> rx.PyDiGraph:
        """Create a rustworkx graph from this node and its children.

        Args:
            with_leaves (bool): True to draw graph with leaves present.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(10).BK()
            >>> rx_graph = tree.root_node.to_rustworkx()
        """
        return to_rustworkx(self, with_leaves=with_leaves)


def update_root_path(root: TTNode, prefix: str) -> None:
    """Prefix the root path of a node and all its children.

    Args:
        root (TTNode): Root ternary tree node.
        prefix (str): String to prefix to root paths.
    """
    logger.debug("Prefixing node root path with %s.", prefix)
    root.root_path = prefix
    for child in ["x", "y", "z"]:
        child_node = getattr(root, child, None)
        if child_node is not None:
            child_node.update_root_path(prefix + child)


def z_descendant(ancestor: TTNode) -> TTNode:
    """Find the furthest z-descendant of a node."""
    node = ancestor
    while isinstance(node.z, TTNode):
        node = node.z
    return node


def z_ancestor(descendant: TTNode) -> TTNode:
    """Find the further z-ancestor of a node."""
    node: TTNode = descendant
    for char in descendant.root_path[::-1]:
        if char != "z":
            break
        node: TTNode = node.parent
    return node


def add_child(
    parent,
    which_child: str,
    child_node: TTNode | None = None,
    qubit_label: int | str | None = None,
) -> TTNode:
    """Add a child node to a parent node.

    Args:
        parent (TTNode): The parent node.
        which_child (str): The child node to add.
        root_path (str): Path from the root node.
        qubit_label (int | str): The qubit label.
        child_node (TTNode | None): A node to assign as child.

    Returns:
        TTNode: The added child node.

    Example:
        >>> from ferrmion.encode.ternary_tree_node import TTNode, add_child
        >>> node = TTNode()
        >>> add_child(node, 'x')
    """
    logger.debug("Adding child %s to parent %s", which_child, parent)

    if (existing_child := getattr(parent, which_child, None)) is not None:
        logger.warning(
            f"Already has child node {existing_child.root_path} at {which_child}"
        )
        return existing_child

    if child_node is None:
        logger.debug("Creating child node.")
        child_node = TTNode()
    elif isinstance(child_node, TTNode) and isinstance(child_node.parent, TTNode):
        logger.warning("Removing child node from current parent.")
        current_position = child_node.root_path[-1]
        setattr(child_node.parent, current_position, None)

    logger.debug("Setting node relationships.")
    try:
        parent.leaf_majorana_indices.pop(which_child)
    except KeyError:
        logger.debug("No leaf index to remove.")

    setattr(parent, which_child, child_node)
    child_node.parent = parent

    child_node.update_root_path(parent.root_path + which_child)

    if qubit_label is not None:
        logger.debug("Replacing child qubit label.")
        child_node.qubit_label = qubit_label

    if which_child == "z":
        child_node.z_ancestor = parent.z_ancestor

    return getattr(parent, which_child)


def as_dict(node: TTNode) -> dict[str, dict]:
    """Create a dictionary of children for a node.

    Args:
        node (TTNode): The node to convert to a dictionary.

    Returns:
        dict[str, dict]: A nested dictionary of children for the node.

    Example:
        >>> from ferrmion.encode.ternary_tree_node import TTNode, as_dict
        >>> node = TTNode()
        >>> node.add_child('x')
        >>> as_dict(node)
    """
    logger.debug("Converting node to dict %s", node)
    children: dict[str, TTNode] = {"x": node.x, "y": node.y, "z": node.z}
    for key, val in children.items():
        if val is not None:
            children[key] = as_dict(children[key])
        else:
            children[key] = node.leaf_majorana_indices[key]
    return children


def child_strings(node: TTNode, prefix: str = "") -> set[str]:
    """Create a list of all child strings for a node.

    Args:
        node (TTNode): The node to convert to a list of strings.
        prefix (str): The prefix for the string.

    Returns:
        list[str]: A list of all child strings for the node.

    Example:
        >>> from ferrmion.encode.ternary_tree_node import TTNode, child_strings
        >>> node = TTNode()
        >>> node.add_child('x')
        >>> child_strings(node)
    """
    logger.debug("Creating child strings for node %s", node)
    strings = {prefix}
    for pauli in ["x", "y", "z"]:
        child = getattr(node, pauli, None)
        if child is not None:
            strings = strings.union(
                child_strings(node=child, prefix=f"{prefix + pauli}")
            )
    logger.debug("Sorting nodes.")
    return strings


def child_qubit_labels(node: TTNode) -> dict[str, int | None]:
    """Return a dict of sorted child nodes and their qubit label.

    Example:
        >>> from ferrmion.encode.ternary_tree_node import TTNode
        >>> node = TTNode()
        >>> node.add_child('x', qubit_label=5)
        >>> node.child_qubit_labels
    """
    label_dict: dict[str, int | str | None] = {}
    for child_string in node.child_strings:
        if child_string == "":
            label_dict[""] = node.qubit_label
            continue

        child: TTNode = node
        for char in child_string:
            # we are using pre-checked strings
            # so we don't need a default
            child = getattr(child, char)
        label_dict[child_string] = child.qubit_label

    return label_dict


def branch_majorana_map(root_node: TTNode) -> dict[str, int]:
    """Create a map from branch strings to majorana indices.

    Args:
        root_node (TTNode): The node to convert to a set of strings.

    Returns:
        dict[str,int]: A map from branch strings to majorana operator indices.

    Example:
        >>> from ferrmion.encode.ternary_tree_node import TTNode, branch_strings
        >>> node = TTNode()
        >>> node.add_child('x')
        >>> branch_majorana_map(node)
    """
    logger.debug("Creating branch strings for node %s", root_node)
    branch_majorana_map = {}
    child_strings = root_node.child_strings
    # possible_leaves = set(*[child+char for char in ["x","y","z"] for child in child_strings])
    # leaves = possible_leaves.difference(child_strings)
    for child in child_strings:
        node = root_node
        for char in child:
            node = getattr(node, char)

        for char in ["x", "y", "z"]:
            if getattr(node, char, None) is None:
                branch_majorana_map[child + char] = node.leaf_majorana_indices[char]

    return branch_majorana_map


def node_sorter(label: str) -> int:
    """This is used to keep the ordring of encodings consistent.

    Args:
        label (str): The label to sort.

    Returns:
        int: Integer label to sort by.

    Example:
        >>> from ferrmion.encode.ternary_tree_node import node_sorter
        >>> node_sorter('xyz')
        123
        >>> node_sorter('xx')
        11
        >>> node_sorter('z')
        3
    """
    if label == "":
        return 0
    pauli_dict = {"x": "1", "y": "2", "z": "3"}
    return int("".join([pauli_dict[item] for item in label.lower()]))


def to_rustworkx(root: TTNode, with_leaves: bool = False) -> rx.PyDiGraph:
    """Convert a TT node and its children to a rustworkx PyDiGraph.

    Args:
        root (TTNode): A node to be the root of the rx graph.
        with_leaves (bool): True to show leaves of tree in graph.

    Example:
        >>> from ferrmion.encode.ternary_tree import TernaryTree
        >>> tree = TernaryTree(10).BK()
        >>> rx_graph = to_rustworkx(tree.root)
    """
    graph = rx.PyDiGraph(check_cycle=True)
    if with_leaves:
        child_dict = {s: i for i, s in enumerate(root.branch_strings)}
    else:
        child_dict = {s: i for i, s in enumerate(root.child_strings)}

    graph.add_nodes_from(child_dict)
    for string in root.child_strings:
        if len(string) == 0:
            continue
        graph.add_edge(child_dict[string[:-1]], child_dict[string], string[-1])
    return graph



import numpy as np
import numpy.typing as npt
from ferrmion.core import anneal_enumerations
from ferrmion.encode import FermionQubitEncoding
from ferrmion.hamiltonians import FermionHamiltonian


def anneal_pauli_weight(
    encoding: FermionQubitEncoding,
    hamiltonian: FermionHamiltonian,
    initial_guess: list[int] | None = None,
    temperature: int | None = None,
) -> tuple[float, npt.NDArray[np.uint]]:
    ipow, sym = encoding._build_symplectic_matrix()
    sigs, coeffs = hamiltonian.signatures_and_coefficients

    if initial_guess is None:
        initial_guess_array = np.arange(hamiltonian.n_modes, dtype=np.uint)
    elif isinstance(initial_guess, list)
        initial_guess_array = np.array([*initial_guess], dtype=np.uint)

    if temperature is None:
        temperature = hamiltonian.n_modes

    return anneal_enumerations(
        ipowers=ipow,
        symplectics=sym,
        signatures=sigs,
        coeffs=coeffs,
        temperature=temperature,
        initial_guess=initial_guess_array,
        coefficient_weighted=False,
    )


def anneal_coefficient_pauli_weight(
    encoding: FermionQubitEncoding,
    hamiltonian: FermionHamiltonian,
    initial_guess: list[int] | None = None,
    temperature: int | None = None,
) -> tuple[float, npt.NDArray[np.uint]]:
    ipow, sym = encoding._build_symplectic_matrix()
    sigs, coeffs = hamiltonian.signatures_and_coefficients

    if initial_guess is None:
        initial_guess_array = np.arange(hamiltonian.n_modes, dtype=np.uint)
    elif isinstance(initial_guess, list)
        initial_guess_array = np.array([*initial_guess], dtype=np.uint)

    if temperature is None:
        temperature = hamiltonian.n_modes

    return anneal_enumerations(
        ipowers=ipow,
        symplectics=sym,
        signatures=sigs,
        coeffs=coeffs,
        temperature=temperature,
        initial_guess=initial_guess_array,
        coefficient_weighted=True,
    )



"""Topology-Preserving Hamiltonian Adaptive Ternary Tree."""

import logging
from itertools import product
from typing import Callable, Iterable

import numpy as np

from ferrmion.encode import TernaryTree
from ferrmion.encode.ternary_tree_node import TTNode
from ferrmion.optimize.hatt import _reduce_hamiltonian

logger = logging.getLogger(__name__)


def _get_node(root_node: TTNode, child_string: str):
    node = root_node
    for char in child_string:
        node = getattr(node, char)
    return node


def _add_all_z_restriction(tree, string_index_map, restrictions):
    child_strings = tree.root_node.child_strings
    all_z = sorted(
        {child for child in child_strings if "x" not in child and "y" not in child},
        key=len,
        reverse=True,
    )[0]
    all_z_index = string_index_map[all_z]
    restrictions[all_z_index][2] = 2 * tree.n_modes
    return restrictions


def _add_retain_child_restrictions(tree: TernaryTree, string_index_map, restrictions):
    """Nodes that have a child."""
    child_strings = tree.root_node.child_strings
    for node in child_strings:
        if node == "":
            continue
        parent = node[:-1]
        parent_index = string_index_map[parent]
        node_index = string_index_map[node]
        match node[-1]:
            case "x":
                restrictions[parent_index][0] = node_index
            case "y":
                restrictions[parent_index][1] = node_index
            case "z":
                restrictions[parent_index][2] = node_index
    return restrictions


def _add_xy_parent_restrictions(
    tree: TernaryTree,
    string_index_map: dict[str, int],
    node_objects_map: dict[int, TTNode],
    restrictions: dict[int, list[int | None]],
):
    """Nodes that have an X-parent or Y-parent."""
    child_strings = tree.root_node.child_strings
    for child in child_strings:
        child_index = string_index_map[child]
        child_node = node_objects_map[child_index]
        # child can be it's own ancestor
        ancestor = child_node.z_ancestor

        if ancestor.root_path == "":
            continue
        # Z-ancestors of an x-child have even z-leaf
        if ancestor.root_path[-1] == "x":
            restrictions[child_index][2] = "Even"
        # Z-ancestors of an y-child have odd z-leaf
        elif ancestor.root_path[-1] == "y":
            restrictions[child_index][2] = "Odd"

    return restrictions


def _get_string_index_map(tree):
    n_leaves = 2 * tree.n_modes + 1
    enumeration_scheme = tree.default_enumeration_scheme()
    child_strings = tree.root_node.child_strings
    return {child: n_leaves + enumeration_scheme[child][1] for child in child_strings}


def _get_node_objects_map(tree, string_index_map):
    return {
        string_index_map[child]: _get_node(tree.root_node, child)
        for child in tree.root_node.child_strings
    }


def _initialise_restrictions(tree: TernaryTree):
    string_index_map = _get_string_index_map(tree)
    node_objects_map = _get_node_objects_map(tree, string_index_map)
    for node in [
        node_objects_map[string_index_map[child]]
        for child in tree.root_node.child_strings
    ]:
        node.leaf_majorana_indices = {k: None for k in node.leaf_majorana_indices}

    # each node has restrictions in tuple (restrictions on x, restrictions on y, restrictions on z)
    restrictions = {i: ["Any", "Any", "Any"] for i in string_index_map.values()}
    restrictions = _add_xy_parent_restrictions(
        tree, string_index_map, node_objects_map, restrictions
    )
    restrictions = _add_retain_child_restrictions(tree, string_index_map, restrictions)
    restrictions = _add_all_z_restriction(tree, string_index_map, restrictions)
    return restrictions


def _unpack_single_restriction(
    restriction: None | list[int] | Callable, unassigned_leaves: set
):
    match restriction:
        case "Any":
            restriction = unassigned_leaves
        case "Even":
            restriction = [i for i in unassigned_leaves if i % 2 == 0]
        case "Odd":
            restriction = [i for i in unassigned_leaves if i % 2 == 1]
        # a REQUIRED assignment will be removed from unassigned_leaves
        case int():
            restriction = [restriction]
        case _:
            raise ValueError("Restriction should be 'Any', 'Even', 'Odd' or int.")
    return restriction


def _update_restrictions(restrictions, parent_node, string_index_map):
    # The node may be its own z-ancestor
    restrictor = parent_node.z_ancestor
    logger.debug(f"{restrictor=}")

    logger.debug(f"{restrictions=}")
    if restrictor.root_path == "":
        pass
    elif restrictor.root_path[-1] == "x":
        if isinstance(restrictor.parent.y, TTNode):
            restricted = restrictor.parent.y.z_descendant
            logger.debug(f"{restricted=}")
            restrictions[string_index_map[restricted.root_path]][2] = (
                restrictor.z_descendant.leaf_majorana_indices["z"] + 1
            )
        elif restrictor.parent.y is None:
            restricted = restrictor.parent
            logger.debug(f"{restricted=}")
            restrictions[string_index_map[restricted.root_path]][1] = (
                restrictor.z_descendant.leaf_majorana_indices["z"] + 1
            )

    elif restrictor.root_path[-1] == "y":
        if isinstance(restrictor.parent.x, TTNode):
            restricted = restrictor.parent.x.z_descendant
            logger.debug(f"{restricted=}")
            restrictions[string_index_map[restricted.root_path]][2] = (
                restrictor.z_descendant.leaf_majorana_indices["z"] - 1
            )
        elif restrictor.parent.x is None:
            restricted = restrictor.parent
            logger.debug(f"{restricted=}")
            restrictions[string_index_map[restricted.root_path]][0] = (
                restrictor.z_descendant.leaf_majorana_indices["z"] - 1
            )
    logger.debug(f"{restrictions=}")
    return restrictions


def _initialise_node_dependencies(tree: TernaryTree):
    string_index_map = _get_string_index_map(tree)
    child_strings = tree.root_node.child_strings
    dependencies = {string_index_map[child]: [] for child in child_strings}

    for child in child_strings:
        for char in ["x", "y", "z"]:
            if child + char in child_strings:
                dependencies[string_index_map[child]].append(
                    string_index_map[child + char]
                )
    return dependencies


def _initialise_active_nodes(index_string_map, node_dependencies):
    active_nodes: dict[int, set[TTNode]] = {}
    for node_index, deps in node_dependencies.items():
        root_len = len(index_string_map[node_index])
        if deps != []:
            continue
        if active_nodes.get(root_len, None) is None:
            active_nodes[root_len] = set()
        active_nodes[root_len].add(node_index)
    return active_nodes


def _update_active_nodes(
    active_nodes, min_parent, index_string_map, node_dependencies, completed_nodes
):
    active_nodes[root_len := len(index_string_map[min_parent])].remove(min_parent)
    if active_nodes[root_len] == set():
        active_nodes.pop(root_len)

    for node, deps in node_dependencies.items():
        if completed_nodes.issuperset(deps):
            root_len = len(index_string_map[node])
            if active_nodes.get(root_len, None) is None:
                active_nodes[root_len] = set()
            active_nodes[root_len].add(node)
    return active_nodes


def _build_valid_combination(
    comb: tuple[int, int],
    descendant_map: dict,
    ancestor_map: dict,
    n_modes: int,
    selection: list[int, int, int],
) -> None | list[int, int, int]:
    # print(f"{comb=}")
    z_index, x_index = comb
    if z_index == x_index:
        return None

    small_y = None
    small_x = None

    small_x = descendant_map[x_index]

    # discard this combination
    if small_x == 2 * n_modes:
        return None

    if small_x % 2 == 0:
        small_y = small_x + 1
    else:
        small_y = small_x - 1
    # We can't use this index for y a
    # it has been used in the combination already
    # so we'd be replacing our x or z!
    if small_y in comb:
        return None

    y_index = ancestor_map[small_y]

    if y_index in comb:
        return None

    if small_x % 2 == 0:
        comb = [x_index, y_index, z_index]
    else:
        comb = [y_index, x_index, z_index]

    # W3 can end uop with the same combination in two different ways
    if comb == selection:
        return None
    return comb


def topphatt(
    majorana_ham: dict[Iterable[int], float],
    tree: TernaryTree,
) -> TernaryTree:
    """Construct an adaptive ternary tree from a majorana Hamiltonian.

    Args:
        majorana_ham (dict[tuple[int,...],float]): Majorana Hamiltonian to encode.
        tree (TernaryTree): The TernaryTree encoding to be optimised
    Returns:
        TernaryTree: Optimised Ternary Tree encoding.
    """
    logger.debug(f"Number of Hamiltonian terms {len(majorana_ham)}")
    n_modes = tree.n_modes
    n_leaves = 2 * n_modes + 1
    # We need 2*M +1 leaves and M nodes.
    nodes: dict[int, TTNode | None] = {i: None for i in range(n_leaves - 1)}
    node_dependencies = _initialise_node_dependencies(tree)  # O(N)
    logging.debug(f"Initial Dependencies:\n{node_dependencies}")

    string_index_map = _get_string_index_map(tree)  # O(N)
    index_string_map = {v: k for k, v in string_index_map.items()}

    node_objects_map = _get_node_objects_map(tree, string_index_map)  # O(N)

    for node in node_objects_map.values():  # O(N)
        for char in ["x", "y", "z"]:
            node.branch_majorana_map[char] = None

    nodes.update(_get_node_objects_map(tree, string_index_map))

    active_nodes: dict[int, set[TTNode]] = _initialise_active_nodes(  # O(N)
        index_string_map=index_string_map, node_dependencies=node_dependencies
    )

    # active_nodes:set[int] = {node for node, deps in node_dependencies.items() if deps == []}
    completed_nodes = set()
    restrictions = _initialise_restrictions(tree)  # O(N)

    logger.debug(f"Initial Restrictions:\n{restrictions}")
    # Start with all the leaves unassigned
    unassigned_leaves = [*range(n_leaves)]
    unassigned_leaves.reverse()

    # We create two maps, of z_ancestors and z_descendants
    ancestor_map = {i: i for i in range(n_leaves + n_modes)}
    descendant_map = {i: i for i in range(n_leaves + n_modes)}

    total_weight = 0
    for i in range(n_modes + 1):  # O(N)
        logging.debug(f"\nLoop {i}")
        # # Update the restrictions with the new information about the tree.
        # # Any nodes that are required to be in a certain position
        # # have to be removed from unassigned!
        to_remove = []
        for restriction in restrictions.values():  # O(N2)
            for term in restriction:
                if isinstance(term, int):
                    to_remove.append(term)
        unassigned_leaves = [l for l in unassigned_leaves if l not in to_remove]
        logger.debug(f"{unassigned_leaves=}")

        min_weight = np.inf
        min_parent = None
        selection = [None, None, None]

        # NOTE
        # Another opion would be to check all allowed
        # combinations just once and then to look
        # for the parent that is eligable for the minimum combination.
        logging.debug(f"{active_nodes=}")
        max_len_active = active_nodes[max([*active_nodes.keys()])]
        logging.debug(f"{max_len_active=}")
        unique_restrictions = set()
        max_len_active = sorted([index_string_map[i] for i in max_len_active])
        for parent_string in max_len_active:  # O ()
            parent_index = string_index_map[parent_string]

            logging.debug(f"\n{parent_index=}")

            # Z-child of new node will always be the previous node.
            # We only need to use every second entry in unassigned
            # as we already order odd/even terms
            # we also know for jw that every new node will be
            # the z-ancestor all all other nodes.

            parent_restrictions = restrictions[parent_index]

            if ("Any", "Any", "Any") in unique_restrictions:
                continue
            elif tuple(parent_restrictions) in unique_restrictions:
                continue
            else:
                unique_restrictions.add(tuple(parent_restrictions))

            logger.debug(f"{unique_restrictions=}")

            allowed_x = _unpack_single_restriction(
                parent_restrictions[0], unassigned_leaves
            )
            allowed_y = _unpack_single_restriction(
                parent_restrictions[1], unassigned_leaves
            )
            allowed_z = _unpack_single_restriction(
                parent_restrictions[2], unassigned_leaves
            )
            logger.debug(f"{allowed_x=}")
            logger.debug(f"{allowed_y=}")
            logger.debug(f"{allowed_z=}")

            match parent_restrictions[0], parent_restrictions[1]:
                case ("Any", "Any"):
                    allowed_product = product(allowed_z, allowed_x)
                case (x, "Any") if x != "Any":
                    allowed_product = product(allowed_z, allowed_x)
                case ("Any", y) if y != "Any":
                    allowed_product = product(allowed_z, allowed_y)
                case (x, y) if "Any" not in [x, y]:
                    allowed_product = product(allowed_x, allowed_y, allowed_z)

            # If x and y are both none, just do all x and create a pair
            # if x is none and y is set, just use y
            # if x is set and y is none, just use x

            logging.debug(f"{parent_restrictions=}")
            for comb in allowed_product:  # O(n choose 2)
                match len(comb):
                    case 2:
                        comb = _build_valid_combination(
                            comb, descendant_map, ancestor_map, n_modes, selection
                        )
                        if comb is None:
                            continue
                    case 3:
                        # We can use the combination provided
                        comb = list(comb)
                    case _:
                        raise ValueError("Length of combination should be 2 or 3.")

                weight = 0
                for key in majorana_ham.keys():  # O(M terms)
                    if min(comb) > max(key):
                        continue
                    elif max(comb) < min(key):
                        continue
                    else:
                        odd_parity_paulis = [
                            sum([t == c for t in key]) % 2
                            for c in comb  # O(k term length)
                        ]
                        non_commuting = sum(odd_parity_paulis) % 3
                        weight += int(non_commuting != 0)
                        if weight > min_weight:
                            break

                if weight < min_weight:
                    logging.debug(
                        f"NEW Min Node:{i}, Parent Index: {parent_index}, Comb: {comb}, Old Min:{min_weight}, New Min:{weight}"
                    )
                    min_weight = weight
                    selection = comb
                    min_parent = parent_index
                # elif weight == min_weight:
                # #     # logging.debug(f"SAME Min Node:{i}, Parent Index: {parent_index}, Comb: {comb}, Old Min:{min_weight }, New Min:{weight}")
                # min_weight = weight
                # selection = comb
                # min_parent = parent_index
            logging.debug(f"{parent_index=}, {selection=}")
        total_weight += min_weight
        logging.debug(f"{min_parent=}")
        # Now find the Y pair of the x-node
        unassigned_leaves = [u for u in unassigned_leaves if u not in selection]
        for child_index, char in zip(selection, ["x", "y", "z"]):
            maybe_node = node_objects_map.get(child_index, None)
            if isinstance(maybe_node, TTNode):
                continue
            node_objects_map[min_parent].leaf_majorana_indices[char] = child_index

        if i + 1 == n_modes:
            break

        logging.debug(
            f"Loop: {i}, Selection: {selection}, Chosen Parent: {min_parent},{index_string_map[min_parent]}"
        )

        completed_nodes.add(min_parent)
        node_dependencies.pop(min_parent)

        active_nodes = _update_active_nodes(
            active_nodes=active_nodes,
            min_parent=min_parent,
            index_string_map=index_string_map,
            node_dependencies=node_dependencies,
            completed_nodes=completed_nodes,
        )

        z_index = selection[2]

        # Update the restrictions on the parent node to be its selection.
        restrictions[min_parent] = selection
        # Use these to update restrictions on other nodes.

        parent_node = node_objects_map[min_parent]
        restrictions = _update_restrictions(
            restrictions, parent_node=parent_node, string_index_map=string_index_map
        )

        z_desc = descendant_map[z_index]
        descendant_map[parent_index] = z_desc
        ancestor_map[z_index] = parent_index
        ancestor_map[z_desc] = parent_index

        majorana_ham = _reduce_hamiltonian(majorana_ham, parent_index, selection)
        logger.debug(f"{majorana_ham=}")

    if len(active_nodes) != 1:
        raise ValueError(f"Not all nodes assigned by HATT. {active_nodes=}")

    tree.pauli_weight = total_weight
    return tree



"""Defines the interface to Qiskit-Nature."""

import logging
from itertools import product

from qiskit.quantum_info import SparsePauliOp
from qiskit_nature.second_q.mappers.fermionic_mapper import FermionicMapper
from qiskit_nature.second_q.operators import FermionicOp

from ferrmion import FermionQubitEncoding
from ferrmion.utils import symplectic_product, symplectic_to_sparse

logger = logging.getLogger(__name__)


class QiskitAdapter(FermionicMapper):
    """Wrapper class enabling the use of ferrmion in qiskit.

    In qiskit_nature, encodings are handled by a general `QubitMapper` class,
    and a `FermionMapper` for Fermion to Qubit encodings.

    These classes have an abstract method `_map_single` which transforms a
    single qiskit operator into a `qiskit.SparsePauliOp`.

    NOTE: You must have installed the optional dependencies with
     `pip install ferrmion[qiskit]` for this functionality to be available.

    Example:
    >>> from ferrmion.encode import JordanWigner
    >>> from ferrmion.interop.qiskit import QiskitAdapter
    >>> from qiskit_nature.second_q.operators import FermionOp
    >>> fop = FermionicOp(
    >>>     {
    >>>         "+_0 -_0": 1.0,
    >>>         "+_1 -_1": -1.0,
    >>>     },
    >>>     num_spin_orbitals=2,
    >>> )
    >>> mapper = QiskitAdapter(JordanWigner(2))
    >>> mapper.map(fop)

    )

    """

    def __init__(self, encoding: FermionQubitEncoding) -> None:
        """Initialise QiskitAdapter.

        Args:
            encoding (FermionQubitEncoding): A valid ferrmion encoding.
        """
        self.encoding = encoding
        super().__init__()

    def _map_single(
        self, second_q_op: FermionicOp, *, register_length: int | None = None
    ) -> SparsePauliOp:
        """Function required to adapt ferrmion encodings to qiskit_nature.

        Allows the use of a ferrmion.FermionQubitEncoding to encode
        qiskit_nature.

        Args:
            second_q_op (qiskit.FermionicOp): A fermionic Operator.
            register_length (int): Number of qubits to use for operator (typically equals the number of modes of encoding).
        """
        if register_length is None:
            register_length = second_q_op.register_length
        ipowers, symplectics = self.encoding._build_symplectic_matrix()

        term_ops = []
        for term, term_coeff in second_q_op.terms():
            majoranas = []
            term_ipowers = []
            logger.debug(f"{term=}, {term_coeff=}")
            for signature, index in term:
                logger.debug(f"{signature=}, {index=}")
                match signature:
                    case "+":
                        term_ipowers.append([0, 3])
                    case "-":
                        term_ipowers.append([0, 1])
                    case _:
                        raise ValueError("Term signature should be + or -")

                majoranas.append([2 * int(index), 2 * int(index) + 1])

            sparse_list = []
            for iterm, comb in zip(product(*term_ipowers), product(*majoranas)):
                logger.debug(f"{comb=}")
                ipower = sum(ipowers[m] for m in comb) + sum(it for it in iterm)
                left = symplectics[comb[0]]
                logger.debug(f"{comb[0]=},{ipower=}, {left=}")
                for m_index in comb[1:]:
                    right = symplectics[m_index]
                    iprod, left = symplectic_product(left, right)
                    ipower += iprod
                    logger.debug(f"{m_index=}, {ipower=}, {left=}")
                sparse = symplectic_to_sparse(left, ipower)
                logger.debug(f"{sparse=}")

                coeff = term_coeff
                coeff *= 0.5 ** len(comb)
                coeff *= sparse[2]

                sparse_list.append((sparse[0], list(sparse[1]), coeff))
                logger.debug(f"{sparse_list=}")
            term_ops.append(
                SparsePauliOp.from_sparse_list(
                    sparse_list, num_qubits=register_length
                ).simplify()
            )
        logger.debug(f"{term_ops=}")

        sparse_op: SparsePauliOp = term_ops[0]
        for term in term_ops[1:]:
            sparse_op += term
        return sparse_op.simplify()



"""Tests for base fermion to qubit encoding class"""

import numpy as np
import pytest
from ferrmion.encode import TernaryTree, MaxNTO
from ferrmion.encode.ternary_tree import JordanWigner, BravyiKitaev, JKMN, ParityEncoding
from ferrmion.encode.base import double_fermionic_operator, FermionQubitEncoding

np.random.seed(1710)


@pytest.fixture
def four_mode_tt():
    return TernaryTree(n_modes=4)


@pytest.fixture
def sixteen_mode_tt():
    return TernaryTree(n_modes=16)


def test_default_vacuum_state(four_mode_tt):
    assert np.all(four_mode_tt.vacuum_state == np.array([0] * 4))


def test_valid_vacuum_state(four_mode_tt):
    with pytest.raises(ValueError) as excinfo:
        four_mode_tt.vacuum_state = [0] * 3
    assert "4" in str(excinfo.value)
    assert "length" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        four_mode_tt.vacuum_state = [0] * 5
    assert "4" in str(excinfo.value)
    assert "length" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        four_mode_tt.vacuum_state = np.array([[0], [0]])
    assert "dimension" in str(excinfo.value)


def test_hartree_fock_state(sixteen_mode_tt):
    jw = sixteen_mode_tt.JW()
    hartree_fock_state = jw.hartree_fock_state
    nq = jw.n_qubits // 2
    print(hartree_fock_state(np.array([True] * nq + [False] * nq, dtype=bool)))
    assert (
        hartree_fock_state(np.array([True] * nq + [False] * nq, dtype=bool))[0]
    ) == [1.0]
    assert np.all(
        hartree_fock_state(np.array([True] * nq + [False] * nq, dtype=bool))[1]
        == np.array([[True] * nq + [False] * nq], dtype=bool)
    )
    assert np.all(
        hartree_fock_state(
            np.array([True] * (nq + 1) + [False] * (nq - 1), dtype=bool)
        )[1]
        == np.array([[True] * (nq + 1) + [False] * (nq - 1)], dtype=bool)
    )


def test_number_operator(four_mode_tt):
    tree = four_mode_tt.JW()
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    # numpy doesn't like comparing empty arrays
    assert str(TernaryTree(n_modes=4).JW().edge_operator((0, 0))) == str(
        TernaryTree(n_modes=4).JW().number_operator(0)
    )
    assert str(TernaryTree(n_modes=4).JW().edge_operator((1, 1))) == str(
        TernaryTree(n_modes=4).JW().number_operator(1)
    )
    assert str(TernaryTree(n_modes=4).JW().edge_operator((2, 2))) == str(
        TernaryTree(n_modes=4).JW().number_operator(2)
    )
    assert str(TernaryTree(n_modes=4).JW().edge_operator((3, 3))) == str(
        TernaryTree(n_modes=4).JW().number_operator(3)
    )

    with pytest.raises(ValueError) as excinfo:
        tree.number_operator(tree.n_modes + 1)
    assert "indices invalid" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        tree.number_operator(-1)
    assert "indices invalid" in str(excinfo.value)


def test_edge_operator(four_mode_tt):
    tree = four_mode_tt.JKMN()
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    left = np.array([t[2] for t in tree.edge_operator((1, 0))], dtype=complex)
    right = np.array(
        [np.conjugate(t[2]) for t in tree.edge_operator((0, 1))], dtype=complex
    )
    assert np.all(right == left[[0, 2, 1, 3]])
    assert np.all(
        left == np.array([0.0 - 0.25j, -0.25 + 0.0j, 0.25 + 0.0j, 0.0 + 0.25j])
    )
    assert np.all(
        right == np.array([0.0 - 0.25j, 0.25 + 0.0j, -0.25 + 0.0j, 0.0 + 0.25j])
    )

    output = tree.edge_operator((0, 3))
    expected = [
        ("XZY", np.array([0, 2, 3]), -0 - 0.25j),
        ("YZY", np.array([0, 2, 3]), 0.25 - 0j),
        ("XZX", np.array([0, 1, 3]), 0.25 + 0j),
        ("YZX", np.array([0, 1, 3]), 0 + 0.25j),
    ]
    for oterm, eterm in zip(output, expected):
        np.all(oitem == eitem for oitem, eitem in zip(oterm, eterm))

    with pytest.raises(ValueError) as excinfo:
        tree.edge_operator((0, tree.n_modes + 1))
    assert "indices invalid" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        tree.edge_operator((tree.n_modes + 1, 0))
    assert "indices invalid" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        tree.number_operator((0, -1))
    assert "indices invalid" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        tree.number_operator((-1, 0))
    assert "indices invalid" in str(excinfo.value)


def test_double_fermionic_operator(four_mode_tt):
    jw_expected = [
        ("", np.array([], dtype=np.int64), 0.25),
        ("Z", np.array([0]), -0.25),
        ("Z", np.array([0]), -0.25),
        ("", np.array([], dtype=np.int64), 0.25),
    ]
    jw_num_zero = double_fermionic_operator(four_mode_tt.JW(), (0, 0), "+-")
    assert jw_num_zero[0][0] == jw_num_zero[3][0] == jw_expected[0][0]
    assert type(jw_num_zero[0][1]) == type(jw_num_zero[3][1]) == type(jw_expected[0][1])
    assert len(jw_num_zero[0][1]) == len(jw_num_zero[3][1]) == len(jw_expected[0][1])
    assert jw_num_zero[0][2] == jw_num_zero[3][2] == jw_expected[0][2]
    assert np.all(jw_num_zero[1] == jw_expected[1])
    assert np.all(jw_num_zero[2] == jw_expected[2])

    bk_expected = [
        ("", np.array([], dtype=np.int64), 0.25),
        ("ZZZ", np.array([0, 1, 3]), -0.25),
        ("ZZZ", np.array([0, 1, 3]), -0.25),
        ("", np.array([], dtype=np.int64), 0.25),
    ]
    bk_num_zero = double_fermionic_operator(four_mode_tt.BK(), (0, 0), "+-")
    assert np.all(l == r for l, r in zip(bk_num_zero[1], bk_expected[1]))
    assert np.all(l == r for l, r in zip(bk_num_zero[2], bk_expected[2]))

    maxnto_expected = [
        ("", np.array([], dtype=np.int64), 0.25),
        ("ZZZ", np.array([0, 1, 2]), 0.25),
        ("ZZZ", np.array([0, 1, 2]), 0.25),
        ("", np.array([], dtype=np.int64), 0.25),
    ]
    maxnot_num_zero = double_fermionic_operator(MaxNTO(4), (0, 0), "+-")
    assert np.all(l == r for l, r in zip(maxnot_num_zero[1], maxnto_expected[1]))
    assert np.all(l == r for l, r in zip(maxnot_num_zero[2], maxnto_expected[2]))


@pytest.mark.parametrize("encoding_func", [JordanWigner, BravyiKitaev, JKMN, ParityEncoding])
def test_majorana_product_doubles_to_idenity(encoding_func):
    encoding: FermionQubitEncoding = encoding_func(5)
    for i in range(5):
        prod, coeff = encoding.majorana_product((i, i))
        assert np.all(prod == np.zeros((2 * 5), dtype=np.bool))
        assert coeff == 1
        assert np.all(
            encoding.majorana_product((i, i, i))[0]
            == encoding.majorana_product((i,))[0]
        )
        assert np.all(
            encoding.majorana_product((i, i, i))[1]
            == encoding.majorana_product((i,))[1]
        )


@pytest.mark.parametrize("encoding_func", [JordanWigner, BravyiKitaev, JKMN, ParityEncoding])
def test_majorana_product_exchange_antisymmetry(encoding_func):
    encoding: FermionQubitEncoding = encoding_func(5)
    for i in range(1, 5):
        assert np.all(
            encoding.majorana_product((i, 0))[0] == encoding.majorana_product((0, i))[0]
        )
        assert np.all(
            encoding.majorana_product((i, 0))[1]
            == -1 * encoding.majorana_product((0, i))[1]
        )


@pytest.mark.parametrize("encoding_func", [JordanWigner, BravyiKitaev, JKMN, ParityEncoding])
def test_majorana_product_empty(encoding_func):
    encoding: FermionQubitEncoding = encoding_func(5)
    assert np.all(encoding.majorana_product(())[0] == np.zeros(10, dtype=bool))
    assert np.all(encoding.majorana_product(())[1] == 1)



"""Huffman-code Ternary Tree."""

import numpy as np
import numpy.typing as npt

from ferrmion.encode import TernaryTree
from ferrmion.encode.ternary_tree_node import TTNode
from ferrmion.utils import find_pauli_weight, pauli_to_symplectic, symplectic_product


def _majarana_op_frequency(
    ones: npt.NDArray[float], twos: npt.NDArray[float]
) -> npt.NDArray[float]:
    n_modes = ones.shape[0]
    majorana_freq = np.zeros(n_modes)

    for i in range(n_modes):
        for j in range(n_modes):
            val = np.abs(ones[i, j])
            positions = {i, j}
            for p in positions:
                majorana_freq[p] += val

    for i in range(n_modes):
        for j in range(n_modes):
            for k in range(n_modes):
                for l in range(n_modes):
                    val = np.abs(twos[i, j, k, l])
                    positions = {i, j, k, l}
                    for p in positions:
                        majorana_freq[p] += val
    return majorana_freq.repeat(2)


def _build_huffman_tree(
    n_modes: int, majorana_frequencies: npt.NDArray[float]
) -> TernaryTree:
    nodes = {i: None for i in range(len(majorana_frequencies))}
    weights = {i: j for i, j in enumerate(majorana_frequencies)}
    n_ops = len(majorana_frequencies)
    for i in range(n_ops // 2):
        parent_index = 2 * n_ops - 1 - i
        mins = sorted(weights.items(), key=lambda kv: (kv[1], kv[0]))[:3]

        parent = nodes.get(parent_index, TTNode(parent=None, qubit_label=i))

        for min, child_string in zip(mins, ["x", "y", "z"][: len(mins)]):
            possible_child = nodes[min[0]]
            if isinstance(possible_child, TTNode):
                parent.add_child(which_child=child_string, child_node=possible_child)

        new_weight = 0
        for index, weight in mins:
            new_weight += weight
            weights.pop(index)
            nodes.pop(index)

        nodes[parent_index] = parent
        weights[parent_index] = new_weight

    root_node = [*nodes.values()][0]
    huffman_tree = TernaryTree(n_modes=n_modes, root_node=root_node)
    huffman_tree.string_pairs

    return huffman_tree


def _two_e_frequency(ones, twos) -> npt.NDArray[float]:
    n_modes = ones.shape[0]
    two_e_freq = np.zeros(ones.shape)
    for j in range(n_modes):
        for i in range(n_modes):
            for l in range(n_modes):
                for k in range(n_modes):
                    val = np.abs(twos[i, j, k, l])
                    two_e_freq[i, j] += val
                    two_e_freq[k, l] += val
    two_e_freq = np.kron(two_e_freq, np.array([[1, 1], [1, 1]]))
    two_e_freq = np.triu(two_e_freq, k=1)
    return two_e_freq


def _mode_priority(two_e_freq):
    vaccum_frequencies = np.diag(two_e_freq, k=1)
    sorted_pairs = [
        (i, i + 1) for i in np.argsort(vaccum_frequencies)[::-1] if i % 2 == 0
    ]
    sorted_modes = [i[0] // 2 for i in sorted_pairs]
    return sorted_modes


def _operator_pair_priority(huffman_tree):
    weights = {}
    for index, pair in enumerate(huffman_tree.string_pairs.values()):
        left, right = pair
        left = huffman_tree.branch_pauli_map[left]
        right = huffman_tree.branch_pauli_map[right]

        weights[index] = {}
        left, _ = pauli_to_symplectic(left, 0)
        right, _ = pauli_to_symplectic(right, 0)
        pair_weight = find_pauli_weight(np.array([left])) + find_pauli_weight(
            np.array([right])
        )
        _, product = symplectic_product(left, right)
        product_weight = find_pauli_weight(np.array([product]))

        weights[index]["pair_weight"] = pair_weight
        weights[index]["prod_weight"] = product_weight

        operator_order = sorted(
            weights.items(), key=lambda kv: (kv[1]["prod_weight"], kv[1]["pair_weight"])
        )

        operator_order = [index for index, _ in operator_order]
    return operator_order


def huffman_ternary_tree(
    one_e_coeffs: npt.NDArray[float], two_e_coeffs: npt.NDArray[float]
) -> TernaryTree:
    """Creates a Huffman-code Ternary Tree.

    Li, Q. S., Liu, H. Y., Wang, Q., Wu, Y. C., & Guo, G. P. (2025).
    Huffman-Code-based Ternary Tree Transformation. Chinese Physics Letters.

    http://iopscience.iop.org/article/10.1088/0256-307X/42/10/100001

    Note: Only vaccum-preserving Huffman-trees are currently supported.

    Args:
        one_e_coeffs (np.ndarray): One electron coefficients.
        two_e_coeffs (np.ndarray): Two electron coefficients.

    Return:
        TernaryTree: A Huffman-code ternary tree.
    """
    n_modes = one_e_coeffs.shape[0]

    majorana_frequencies = _majarana_op_frequency(one_e_coeffs, two_e_coeffs)

    huffman_ternary_tree = _build_huffman_tree(n_modes, majorana_frequencies)
    huffman_ternary_tree.enumeration_scheme = (
        huffman_ternary_tree.default_enumeration_scheme()
    )

    two_e_frequencies = _two_e_frequency(one_e_coeffs, two_e_coeffs)
    sorted_modes = _mode_priority(two_e_frequencies)
    sorted_operators = _operator_pair_priority(huffman_ternary_tree)

    mode_op_map = [0] * len(sorted_modes)
    for operator_index, mode_index in enumerate(sorted_modes):
        mode_op_map[mode_index] = sorted_operators[operator_index]

    huffman_ternary_tree.default_mode_op_map = mode_op_map
    huffman_ternary_tree.enumeration_scheme = (
        huffman_ternary_tree.default_enumeration_scheme()
    )
    return huffman_ternary_tree



"""Reduced entanglement Ternary Tree."""

import logging

import numpy as np
from numpy.typing import NDArray

from ..encode.ternary_tree import TernaryTree
from ..encode.ternary_tree_node import TTNode

logger = logging.getLogger(__name__)


def reduced_entanglement_ternary_tree(
    mutual_information: NDArray,
    cutoff: float = 0.5,
    max_branches: int | None = None,
    squash: bool = False,
) -> TernaryTree:
    """Creates the reduced entanglement TernaryTree.

    Args:
        mutual_information (NDArray): A 2D array of mode mutual information.
        cutoff (float | None): The average MI between spatial orbitals.
        max_branches (int): The maximum allowed number of Parity branches.
        squash (bool): Whether to squash the mutual_information from spin-orbit form to spinless.

    Returns:
        TernaryTree: A new ternary tree.

    Note:
        Assumes that the MI matrix gives MI between spinless spatial orbitals
        So that each block of four contains [[aa, ab], [ba,bb]]

    Example:
        >>> import numpy as np
        >>> from ferrmion.optimize.rett import reduced_entanglement_tree
        >>> mi = 0.5 * np.random.random((6,6))
        >>> mi = mi + mi.T
        >>> tree = reduced_entanglement_tree(mi)
        >>> tree.as_dict()

    Advanced example (with options):
        >>> tree = reduced_entanglement_tree(mi, cutoff=0.1, max_branches=2, squash=False)
    """
    logger.debug("Creating Reduced entanglement TT.")
    enumeration_scheme = {}
    n_modes = mutual_information.shape[0]
    n_modes *= 1 if squash else 2

    new_tree = TernaryTree(n_modes, root_node=TTNode())

    if squash:
        # First combine the MI information for alpha and beta spins
        squash_rows = mutual_information[::2] + mutual_information[1::2]
        squash_matrix = squash_rows[:, ::2] + squash_rows[:, 1::2]
        squash_matrix *= 0.25
    else:
        squash_matrix = mutual_information

    mi_rank = np.triu(squash_matrix).flatten().argsort()[::-1]
    # Convert back to square format from flattened
    sorted_indices = [np.unravel_index(index, squash_matrix.shape) for index in mi_rank]
    sorted_indices = [(int(i[0]), int(i[1])) for i in sorted_indices]
    logger.debug(f"Matrix indices sorted by decreasing MI: {sorted_indices}")

    branches: list[tuple[int, int, int, int]] = []
    unused_indices = {i for i in range(squash_matrix.shape[0])}
    for squash_index in sorted_indices:
        if len(set(squash_index)) == 1:
            logger.warning("MI Matrix contains non-zero diagonal elements, skipping.")
            continue

        if max_branches is not None and len(branches) >= max_branches:
            break

        if not unused_indices.issuperset(squash_index):
            logger.debug("Indices %s previously assigned to branch.", squash_index)
            continue

        if squash_matrix[squash_index] >= cutoff:
            branch = (
                2 * squash_index[0],
                2 * squash_index[0] + 1,
                2 * squash_index[1],
                2 * squash_index[1] + 1,
            )
            logger.debug("Adding branch %s", branch)
            unused_indices.remove(squash_index[0])
            unused_indices.remove(squash_index[1])
            branches.append(branch)

        if len(unused_indices) <= 1:
            break

    unused_modes = {i for i in range(new_tree.n_qubits)}
    for i, branch in enumerate(branches):
        for j, mode in enumerate(branch):
            node_path = "z" * i + "x" * j
            new_tree.add_node(node_path)
            enumeration_scheme[node_path] = (mode, mode)
            unused_modes.remove(mode)

    remaining_modes = new_tree.n_qubits - (4 * len(branches))
    new_tree.add_node("z" * (remaining_modes + len(branches) - 1))

    for node_path in new_tree.root_node.child_strings:
        if enumeration_scheme.get(node_path, None) is None:
            mode = unused_modes.pop()
            enumeration_scheme[node_path] = (mode, mode)

    logger.debug("Setting enumeration scheme")
    logger.debug(enumeration_scheme)
    new_tree.enumeration_scheme = enumeration_scheme
    return new_tree



"""Init for Qiskit Interop.

These functions require ferrmion is installed with optional extra dependencies
for example, using pip: `pip install ferrmion[qiskit]`
"""

try:
    from qiskit_mapper import QiskitAdapter

    __all__ = ["QiskitAdapter"]
except ImportError:
    __all__ = []



"""Utility functions."""

import datetime
import json
import logging
import logging.config
from itertools import product

import numpy as np
from numpy.typing import NDArray

from ferrmion.core import symplectic_product

logger = logging.getLogger(__name__)


def icount_to_sign(icount: int) -> np.complex64:
    """Convert a power of i to a complex value.

    Args:
        icount (int): The power of i.

    Returns:
        np.complex64: The complex value.

    Example:
        >>> from ferrmion.utils import icount_to_sign
        >>> icount_to_sign(1)
        1j
        >>> icount_to_sign(2)
        -1
    """
    vals = {0: 1, 1: 1j, 2: -1, 3: -1j}
    return vals[icount % 4]


def symplectic_hash(symp: NDArray[bool]) -> bytes:
    """Convert a symplectic vector into a hashable form.

    Args:
        symp (NDArray[bool]): The symplectic vector.

    Returns:
        bytes: The hashed form of the symplectic vector.

    Example:
        >>> import numpy as np
        >>> from ferrmion.utils import symplectic_hash
        >>> symp = np.array([True, False, True, False], dtype=bool)
        >>> h = symplectic_hash(symp)
        >>> isinstance(h, bytes)
        True
    """
    return np.packbits(symp).tobytes()


def symplectic_unhash(symp: bytes, length: int) -> NDArray[bool]:
    """Convert a hashed symplectic vector back to its original form.

    Args:
        symp (bytes): The hashed form of the symplectic vector.
        length (int): The length of the original symplectic vector.

    Returns:
        NDArray[bool]: The original symplectic vector.

    Example:
        >>> import numpy as np
        >>> from ferrmion.utils import symplectic_hash, symplectic_unhash
        >>> arr = np.array([True, False, True, False], dtype=bool)
        >>> h = symplectic_hash(arr)
        >>> arr2 = symplectic_unhash(h, 4)
        >>> np.all(arr == arr2)
        True
    """
    unpacked = np.unpackbits(np.frombuffer(symp, dtype=np.uint8))
    if len(unpacked) < length:
        unpacked = np.pad(
            unpacked, length - len(unpacked), "constant", constant_values=0
        )
    return np.array(unpacked[:length], dtype=bool)


def symplectic_to_pauli(
    symplectic: NDArray[bool],
    ipower: int = 0,
) -> tuple[str, int]:
    """Convert a symplectic vector into a Pauli String.

    Args:
        symplectic (NDArray[np.uint8]) : symplectic vector [X terms, Y terms]
        ipower (NDArray[np.uint]): power of i coefficient

    Returns:
        tuple[str, int]: The Pauli string and imaginary cofactor.

    NOTE: symplectic XZ does represent XZ and not Y
        So Y=-iXZ needs an imaginary cofactor

    Example:
        >>> import numpy as np
        >>> from ferrmion.utils import symplectic_to_pauli
        >>> arr = np.array([1, 0, 0, 1], dtype=bool)
        >>> pauli, ipower = symplectic_to_pauli(arr)
        >>> isinstance(pauli, str)
        True
    """
    left, right = np.hsplit(symplectic, 2)
    total = left + 2 * right

    def to_pauli(x):
        match x:
            case 0:
                return "I"
            case 1:
                return "X"
            case 2:
                return "Z"
            case 3:
                return "Y"

    to_paulis = np.vectorize(to_pauli)
    pauli_list = to_paulis(total)

    pauli_string = "".join(pauli_list)
    y_count = pauli_string.count("Y")
    ipower += 3 * y_count
    ipower %= 4
    return pauli_string, ipower


def symplectic_to_sparse(
    symplectic: NDArray[bool],
    ipower: int = 0,
) -> tuple[str, NDArray[int], np.complex64]:
    """Convert a symplectic vector into a Pauli String (sparse form).

    Args:
        symplectic (NDArray[np.uint8]) : symplectic vector [X terms, Y terms]
        ipower (NDArray[np.uint]): power of i coefficient

    Returns:
        tuple[str, NDArray[int]]: The Pauli string, indices of non-identity terms and imaginary coefficient.

    NOTE: symplectic XZ does represent XZ and not Y
        So Y=-iXZ needs an imaginary cofactor

    Example:
        >>> import numpy as np
        >>> from ferrmion.utils import symplectic_to_sparse
        >>> arr = np.array([1, 0, 0, 1], dtype=bool)
        >>> pauli, idx, coeff = symplectic_to_sparse(arr)
        >>> isinstance(pauli, str)
        True
        >>> isinstance(idx, np.ndarray)
        True
    """
    xhalf, zhalf = np.hsplit(symplectic, 2)
    total = xhalf + 2 * zhalf

    def to_pauli(x):
        match x:
            case 0:
                return ""
            case 1:
                return "X"
            case 2:
                return "Z"
            case 3:
                return "Y"

    to_paulis = np.vectorize(to_pauli)
    pauli_list = to_paulis(total)

    pauli_string = "".join(pauli_list)
    indices = np.where(total != 0)[0]
    y_count = pauli_string.count("Y")
    ipower += 3 * y_count
    ipower %= 4
    return pauli_string, indices, icount_to_sign(ipower)


def pauli_to_symplectic(
    pauli: str,
    ipower: int = 0,
) -> tuple[NDArray[bool], int]:
    """Convert a Pauli operator to symplectic form.

    Args:
        pauli (str): The Pauli operator string.
        ipower (NDArray[np.uint]): power of i coefficient

    Returns:
        tuple[int, NDArray[np.uint8, np.uint8]]: The imaginary cofactor and symplectic matrix.

    Example:
        >>> from ferrmion.utils import pauli_to_symplectic
        >>> symp, ipower = pauli_to_symplectic('XIZY')

    """
    pauli_array = np.array(list(pauli))
    x_map = {
        "I": 0,
        "X": 1,
        "Y": 1,
        "Z": 0,
    }
    z_map = {
        "I": 0,
        "X": 0,
        "Y": 1,
        "Z": 1,
    }
    # each y is turned into a iY=XZ
    y_count = np.count_nonzero(pauli_array == "Y") % 4
    ipower += y_count
    ipower %= 4
    # logger.debug(f{y_count=})
    x_array = np.array([x_map[term] for term in pauli], dtype=bool)
    z_array = np.array([z_map[term] for term in pauli], dtype=bool)
    return np.hstack((x_array, z_array), dtype=bool), ipower


def xz_swap(symplectic) -> NDArray[bool]:
    """Swap X and Z Pauli operators in a symplectic matrix.

    Args:
        symplectic (NDArray): The symplectic matrix.

    Returns:
        NDArray[np.uint8]: The symplectic matrix with X and Z swapped.

    Example:
        >>> import numpy as np
        >>> from ferrmion.utils import xz_swap
        >>> arr = np.zeros((2, 4), dtype=bool)
        >>> swapped = xz_swap(arr)

    """
    logger.debug(f"Swapping X and Z in symplectic matrix\n{symplectic=}")
    symplectic = np.array(symplectic, dtype=np.uint8)
    x_block, z_block = np.hsplit(symplectic, 2)
    is_z = np.where(np.logical_and(z_block, np.logical_not(x_block)))
    is_x = np.where(np.logical_and(x_block, np.logical_not(z_block)))

    new_x_block = np.copy(x_block)
    new_x_block[is_z] = True
    new_x_block[is_x] = False

    new_z_block = np.copy(z_block)
    new_z_block[is_x] = True
    new_z_block[is_z] = False
    return np.hstack((new_x_block, new_z_block))


def xy_swap(symplectic) -> NDArray[np.uint8]:
    """Swap X and Y Pauli operators in a symplectic matrix.

    Args:
        symplectic (NDArray): The symplectic matrix.

    Returns:
        NDArray[np.uint8]: The symplectic matrix with X and Y swapped.

    Example:
        >>> import numpy as np
        >>> from ferrmion.utils import xy_swap
        >>> arr = np.zeros((2, 4), dtype=bool)
        >>> swapped = xy_swap(arr)

    """
    logger.debug(f"Swapping X and Y in symplectic matrix\n{symplectic=}")
    symplectic = np.array(symplectic, dtype=np.uint8)
    x_block, z_block = np.hsplit(symplectic, 2)
    is_y = np.where(x_block + z_block == 2)
    is_x = np.where(x_block - z_block == 1)

    new_x_block = np.copy(x_block)
    new_x_block[is_x] = 1
    new_x_block[is_y] = 1

    new_z_block = np.copy(z_block)
    new_z_block[is_x] = 1
    new_z_block[is_y] = 0
    return np.hstack((new_x_block, new_z_block))


def yz_swap(symplectic) -> NDArray[np.uint8]:
    """Swap Y and Z Pauli operators in a symplectic matrix.

    Args:
        symplectic (NDArray): The symplectic matrix.

    Returns:
        NDArray[np.uint8]: The symplectic matrix with Y and Z swapped.

    Example:
        >>> import numpy as np
        >>> from ferrmion.utils import yz_swap
        >>> arr = np.zeros((2, 4), dtype=bool)
        >>> swapped = yz_swap(arr)

    """
    symplectic = np.array(symplectic, dtype=np.uint8)
    x_block, z_block = np.hsplit(symplectic, 2)
    is_y = np.where(x_block + z_block == 2)
    is_z = np.where(z_block - x_block == 1)

    new_x_block = np.copy(x_block)
    new_x_block[is_y] = 0
    new_x_block[is_z] = 1

    new_z_block = np.copy(z_block)
    new_z_block[is_y] = 1
    new_z_block[is_z] = 1
    return np.hstack((new_x_block, new_z_block))


def qubit_swap(symplectic, index_pair) -> NDArray[np.uint8]:
    """Swap the position of two qubits in a symplectic matrix.

    Args:
        symplectic (NDArray): The symplectic matrix.
        index_pair (tuple[int]): The indices of the qubits to swap.

    Returns:
        NDArray[np.uint8]: The symplectic matrix with the qubits swapped.

    Example:
        >>> import numpy as np
        >>> from ferrmion.utils import qubit_swap
        >>> arr = np.zeros((2, 4), dtype=bool)
        >>> swapped = qubit_swap(arr, (0, 1))

    """
    logger.debug(f"Swapping qubits {index_pair} in symplectic matrix\n{symplectic=}")
    half_length = symplectic.shape[1] // 2
    i1, i2 = index_pair
    x1 = np.copy(symplectic[:, i1])
    x2 = np.copy(symplectic[:, i2])
    z1 = np.copy(symplectic[:, half_length + i1])
    z2 = np.copy(symplectic[:, half_length + i2])
    symplectic[:, i2] = x1
    symplectic[:, i1] = x2
    symplectic[:, half_length + i1] = z2
    symplectic[:, half_length + i2] = z1
    return symplectic


def check_trivial_overlap(symplectic) -> tuple[bool, NDArray[np.uint]]:
    """Check the Non-trivial Overlap of a symplectic matrix.

    Args:
        symplectic (NDArray): The symplectic matrix.

    Returns:
        tuple[bool, NDArray[int]]: A boolean indicating if the overlap is trivial and the overlap matrix.

    Example:
        >>> import numpy as np
        >>> from ferrmion.utils import check_trivial_overlap
        >>> arr = np.eye((4, 4), dtype=bool)
        >>> satisfied, nto = check_trivial_overlap(arr)
        >>> isinstance(satisfied, bool)
        True
    """
    symplectic = np.array(symplectic, dtype=np.uint8)
    logger.debug(f"Checking trivial overlap\n{symplectic=}")
    x_length = int(len(symplectic[0]) / 2)

    symp_x = symplectic[:, :x_length]
    symp_z = symplectic[:, x_length:]
    symp_i = np.abs(symp_x - 1) * np.abs(symp_z - 1)
    symp_y = symp_x * symp_z

    symp_x = symp_x - symp_y
    symp_z = symp_z - symp_y

    i_trivial = symp_i @ symp_i.T
    same_p_trivial = symp_x @ symp_x.T + symp_y @ symp_y.T + symp_z @ symp_z.T
    one_i_trivial = (
        symp_x @ symp_i.T
        + symp_i @ symp_x.T
        + symp_y @ symp_i.T
        + symp_i @ symp_y.T
        + symp_z @ symp_i.T
        + symp_i @ symp_z.T
    )
    all_trivial = i_trivial + same_p_trivial + one_i_trivial

    nto: NDArray[np.uint] = all_trivial.shape[0] / 2 - all_trivial

    satisfied: bool = np.all((nto + np.eye(nto.shape[0])) % 2 == 1)

    logger.debug(f"Trivial overlap satisfied: {satisfied}")
    logger.debug(f"Trivial overlap matrix:\n{nto}")
    return satisfied, nto


def two_operator_product(creation: tuple[bool, bool], left, right) -> NDArray:
    """Calculate the product of two operators in symplectic form.

    Args:
        creation (tuple[bool, bool]): A tuple of two booleans indicating if the operators are creation operators.
        left (NDArray): The left operator in symplectic form.
        right (NDArray): The right operator in symplectic form.

    Returns:
        NDArray: The product of the two operators in symplectic form.

    Example:
        >>> left = np.array([[1, 0], [0, 1]])
        >>> right = np.array([[0, 1], [1, 0]])
        >>> creation = (True, False)
        >>> two_operator_product(creation, left, right)
            array([
                [0, 0],
                [1, 0]
                [0, 1],
                [0, 0]])
    """
    logger.debug("Calculating two operator product.")
    # (a+ib)(c+id) -> ac, iad, ibc, -bd
    first_term = symplectic_product(left[:, 0], right[:, 0])
    second_term = symplectic_product(left[:, 0], right[:, 1])
    third_term = symplectic_product(left[:, 1], right[:, 0])
    fourth_term = symplectic_product(left[:, 1], right[:, 1])

    # left creation -> -iad, +bd
    # right creation -> -ibc, +bd
    # both creation -> -iad, -ibc, -bd
    if creation[0] is True:
        second_term[0] += 2
        fourth_term[0] += 2
    if creation[1] is True:
        third_term[0] += 2
        fourth_term[0] += 2

    return np.vstack((first_term, second_term, third_term, fourth_term))


def find_pauli_weight(symplectic_hamiltonian: NDArray[bool]) -> np.floating:
    """Find the average Pauli weight of a symplectic hamiltonian.

    Args:
        symplectic_hamiltonian (NDArray): The symplectic Hamiltonian.

    Returns:
        float: The average Pauli weight.

    Example:
        >>> import numpy as np
        >>> from ferrmion.utils import find_pauli_weight
        >>> arr = np.eye((2, 4), dtype=bool)
        >>> find_pauli_weight(arr)
        1.0
    """
    logger.debug("Finding Pauli weight of symplectic Hamiltonian")
    half_length = symplectic_hamiltonian.shape[-1] // 2
    has_pauli = np.bitwise_or(
        symplectic_hamiltonian[:, :half_length], symplectic_hamiltonian[:, half_length:]
    )
    return np.mean(np.sum(has_pauli, axis=1))


def save_pauli_ham(
    pauli_hamiltonian: dict[str, float], filename: str | None = None
) -> None:
    """Save the Pauli Hamiltonian to a JSON file.

    Args:
        pauli_hamiltonian (dict[str, float]): The Pauli Hamiltonian.
        filename (str, optional): The filename (without extension) to save the Hamiltonian to. Defaults to None.

    Example:
        >>> from ferrmion.utils import save_pauli_ham
        >>> save_pauli_ham({'X': 1.0, 'Z': -1.0}, filename='test_ham')
    """
    logger.debug("Saving Pauli Hamiltonian to JSON file")
    if filename is None:
        filename = "pauli_hamiltonian_" + str(datetime.datetime.now())

    with open(f"{filename}.json", "w") as f:
        f.write(json.dumps(pauli_hamiltonian))
    logger.debug(f"Saved Pauli Hamiltonian to {filename}")


def _signature_char_to_ipowers(char: str) -> list:
    """Convert a signature character to a list of imaginary factors.

    Args:
        char(str): One of + or -

    Returns:
        list: A length-2 list of imaginary factors.
    """
    match char:
        case "+":
            result = [1, -1j]
        case "-":
            result = [1, 1j]
        case "_":
            raise ValueError("Signature must contain only + and -.")
    return result


def _hamiltonian_term_to_majorana(
    majorana_ham: dict[tuple[int, ...], float],
    coeffs: np.typing.NDArray[float],
    signature: str,
) -> dict[tuple[int, ...], float]:
    """Add a fermionic Hamiltonian term to a sparse majorana Hamiltonian.

    Args:
        majorana_ham (dict): The majorana Hamiltonian to update.
        coeffs (np.ndarray): The coefficients of the fermionic term.
        signature (str): The signature of the fermionic term.

    Returns:
        dict: The updated majorana Hamiltonian.

    """
    assert len(signature) == coeffs.ndim
    non_zero = np.where(coeffs != 0)
    non_zero_ones = [(*indices, coeffs[indices]) for indices in zip(*non_zero)]
    normalisation = 0.5 ** len(signature)

    ipowers = np.array([_signature_char_to_ipowers(c) for c in signature])
    for *inds, coeff in non_zero_ones:
        # we need two majoranas for each fermionic operator
        # 0 -> left, 1 -> right
        left_right_indices = [[0, 1]] * len(signature)
        for left_right in product(*left_right_indices):
            majorana_ind = [int(2 * i + lr) for i, lr in zip(inds, left_right)]

            swaps = 0
            n = len(majorana_ind)
            for i in range(n - 1):
                swapped = False
                for j in range(n - i - 1):
                    if majorana_ind[j] > majorana_ind[j + 1]:
                        majorana_ind[j], majorana_ind[j + 1] = (
                            majorana_ind[j + 1],
                            majorana_ind[j],
                        )
                        swapped = True
                        swaps += 1
                if not swapped:
                    break
            no_duplicates = []
            for ind in range(0, len(majorana_ind)):
                if majorana_ind.count(majorana_ind[ind]) % 2 == 1:
                    no_duplicates.append(majorana_ind[ind])

            if no_duplicates == []:
                continue

            majoranas = tuple(no_duplicates)
            term_ipowers = np.prod([ipow[lr] for ipow, lr in zip(ipowers, left_right)])
            term_ipowers = term_ipowers * ((-1) ** (swaps % 2))
            majorana_ham[majoranas] = majorana_ham.get(majoranas, 0)
            majorana_ham[majoranas] += normalisation * coeff * term_ipowers
    return majorana_ham


def fermionic_to_sparse_majorana(
    hamiltonian_terms: list[tuple[np.ndarray, str]],
) -> dict[tuple[int], np.complex64]:
    """Convert a list of fermionic Hamiltonian terms to a sparse majorana Hamiltonian.

    Args:
        hamiltonian_terms (list): A list of tuples, each containing a numpy array of coefficients and a signature string.

    Returns:
        dict: A sparse majorana Hamiltonian, with majorana indices as keys and coefficients as values.

    Example:
        >>> from openfermionpyscf import *
        >>> from openfermion import spinorb_from_spatial

        >>> mol="H2O"
        >>> geometry = geometry_from_pubchem(mol)
        >>> basis = "sto-3g"
        >>> multiplicity = 1
        >>> charge = 0

        >>> molecule = MolecularData(geometry, basis, multiplicity, charge)
        >>> molecule = run_pyscf(molecule,run_scf=True,run_fci=False) # NOTE: running FCI is expensive! Don't try on large systems
        >>> ones = molecule.one_body_integrals
        >>> twos = molecule.two_body_integrals

        >>> ones,twos = spinorb_from_spatial(ones, twos)
        >>> twos = 0.5*twos

        >>> # For the molecular Hamiltonian in physicist notation
        >>> majorana_ham = fermionic_to_sparse_majorana([(ones, "+-"), (twos,"++--")])
    """
    total_ham: dict = {}
    for coeffs, signature in hamiltonian_terms:
        total_ham.update(
            _hamiltonian_term_to_majorana(total_ham, coeffs=coeffs, signature=signature)
        )

    return {k: v for k, v in total_ham.items() if np.abs(v) > 1e-16}


def setup_logs() -> None:
    """Initialise logging."""
    config_dict = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {
                "format": "%(asctime)s: %(name)s: %(lineno)d: %(levelname)s: %(message)s"
            },
        },
        "handlers": {
            "file_handler": {
                "class": "logging.FileHandler",
                "level": "DEBUG",
                "formatter": "standard",
                "filename": ".ferrmion.log",
                "mode": "w",
                "encoding": "utf-8",
            },
            "stream_handler": {
                "class": "logging.StreamHandler",
                "level": "WARNING",
                "formatter": "standard",
            },
        },
        "loggers": {
            "": {"handlers": ["file_handler", "stream_handler"], "level": "DEBUG"}
        },
    }

    logging.config.dictConfig(config_dict)
    logger = logging.getLogger(__name__)
    logger.debug("Logging initialised.")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

# sys.path.insert(0, os.path.abspath("../../python/ferrmion"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'ferrmion'
copyright = '2025, Michael Williams de la Bastida'
author = 'Michael Williams de la Bastida'
version = "0.5.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.autodoc",
    # "sphinx.ext.doctest",
    # "sphinx.ext.todo",
    "sphinx.ext.mathjax",
    # "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    # "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "myst_nb",
    "pydata_sphinx_theme",
    "myst_sphinx_gallery",
    ]

myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_image",
]

nb_scroll_outputs = True
nb_execution_timeout = 600

templates_path = ['_templates']
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'
html_static_path = ['_static']
html_theme_options = {
    "github_url":"https://github.com/UCL-CCS/ferrmion",
}



"""Class and methods to easily build general Fermion Hamiltonians."""

import logging

import numpy as np
import numpy.typing as npt
from numpy.typing import NDArray

logger = logging.getLogger(__name__)

"""
Type alias for qubit hamiltonians.
"""
type QubitHamiltonian = dict[str, float]


class FermionHamiltonian:
    """Class for building Fermionic Hamiltonians."""

    def __init__(self, *, terms: dict[str, NDArray] = {}, constant_energy: float = 0.0):
        logger.debug("Initialising FermionHamiltonian")
        self._terms: dict[str, NDArray] = terms
        self.constant_energy = constant_energy
        self._next_term = ""
        self.n_modes = 0
        self._check_and_set_n_modes()

    def _check_and_set_n_modes(self):
        if len(self._terms) == 0:
            pass
        elif len(self._terms) > 0 and self.n_modes == 0:
            logger.debug(f"Setting n_modes: {self.n_modes}")
            self.n_modes = [*self._terms.values()][0].shape[0]
        else:
            for term in self._terms.values():
                if np.any([side != self.n_modes for side in term.shape]):
                    logger.error(
                        f"Hamiltonian coefficient {term.shape} must have constant length on all dimensions."
                    )
                    raise ValueError(
                        f"Hamiltonian coefficient {term.shape} must have constant length on all dimensions."
                    )

    def __repr__(self):
        terms = ", ".join([*self._terms])
        return f"FermionHamiltonian({terms}, {self.n_modes} modes, constant {self.constant_energy})"

    @property
    def signatures_and_coefficients(self) -> tuple[list[str], list[NDArray]]:
        sigs: list[str] = []
        coeffs: list[NDArray] = []
        for k, v in self._terms.items():
            sigs.append(k)
            coeffs.append(v)

        return (sigs, coeffs)

    def creation(self) -> "FermionHamiltonian":
        self._next_term += "+"
        return self

    def annihilation(self) -> "FermionHamiltonian":
        self._next_term += "-"
        return self

    def with_coefficients(self, coefficients: NDArray) -> "FermionHamiltonian":
        if coefficients.ndim != len(self._next_term):
            logger.error(f"Cannot apply coefficents to term {self._next_term}")
        else:
            self._terms[self._next_term] = coefficients
            self._next_term = ""
            self._check_and_set_n_modes()
        return self

    def add_constant(self, constant_energy: float) -> "FermionHamiltonian":
        self.constant_energy += constant_energy
        return self


def molecular_hamiltonian(
    one_e_coeffs: NDArray,
    two_e_coeffs: NDArray,
    constant_energy: float = 0.0,
    physicist_notation: bool = True,
) -> FermionHamiltonian:
    """Return an encoded electronic stucture hamiltonain with niave enumeration.

    Args:
        encoding (FermionQubitEncoding): The encoding to use.
        one_e_coeffs (NDArray): One electron hamiltonian coefficients in spinorb format.
        two_e_coeffs (NDArray): Two electron hamiltonian coefficients in spinorb format.
        constant_energy (float): Constant energy offset.
        physicist_notation (bool): Set to False for Chemist Notation.

    Example:
        >>> import numpy as np
        >>> from ferrmion.hamiltonians.molecular import molecular_hamiltonian
        >>> from ferrmion.encode import TernaryTree
        >>> tree = TernaryTree(12).JW()
        >>> one_e = np.eye((2,2))
        >>> two_e = np.eye((2,2,2,2))
        >>> fham = molecular_hamiltonian(one_e, two_e, 0.0)
        >>> tree.encode(fham)
    """
    if physicist_notation:
        terms = {"+-": one_e_coeffs, "++--": two_e_coeffs}
    else:
        terms = {"+-": one_e_coeffs, "+-+-": two_e_coeffs}

    return FermionHamiltonian(terms=terms, constant_energy=constant_energy)


def linear_adjacency_matrix(length: int, periodic: bool) -> npt.NDArray[bool]:
    """Creates an adjacency matrix for a linear Hubbard Hamiltonian.

    Args:
        length (int): The number of sites.
        periodic (bool): If true, periodic boundary conditions are used.

    Returns:
        np.ndarray[bool]: Adjacency matrix for lattice sites.
    """
    return square_lattice_adjacency_matrix((length, 1), periodic=periodic)


def square_lattice_adjacency_matrix(
    shape: tuple[int, int], periodic: bool
) -> npt.NDArray[bool]:
    """Creates an adjacency matrix for a 2D square lattice Hubbard Hamiltonian.

    Args:
        shape (tuple[int, int]): The number of sites.
        periodic (bool): If true, periodic boundary conditions are used.

    Returns:
        np.ndarray[bool]: Adjacency matrix for lattice sites.
    """
    # find the side length to fit nodes into square
    # we'll build a perfect square first before cutting.
    nx, ny = shape
    n_sites = nx * ny

    # initially make a chain
    adjacency_matrix = np.eye(n_sites, k=1)

    # cut chain into rows by removing connections
    for i in range(nx, n_sites, nx):
        adjacency_matrix[i - 1, i] = 0.0

    # Add connection to number below.
    adjacency_matrix += np.eye(n_sites, k=nx)

    if periodic:
        # Wrap rows
        for i in range(ny):
            adjacency_matrix[i * nx, (i + 1) * nx - 1] = 1

        # Wrap columns
        adjacency_matrix += np.eye(n_sites, k=nx * (ny - 1))

    # Hamitian conjugate
    adjacency_matrix += adjacency_matrix.T
    return np.array(adjacency_matrix, dtype=bool)


def cube_lattice_adjacency_matrix(
    shape: tuple[int, int, int], periodic: bool
) -> npt.NDArray[bool]:
    """Creates an adjacency matrix for a 3D square lattice Hubbard Hamiltonian.

    Args:
        shape (tuple[int, int, int]): The number of sites.
        periodic (bool): If true, periodic boundary conditions are used.

    Returns:
        np.ndarray[bool]: Adjacency matrix for lattice sites.
    """
    nx, ny, nz = shape
    n_sites = nx * ny * nz

    adjacency_matrix = np.zeros((n_sites, n_sites))
    # Add each of the layers of a square matrix
    for i in range(0, n_sites, nx * ny):
        adjacency_matrix[i : i + nx * ny, i : i + nx * ny] = np.triu(
            square_lattice_adjacency_matrix((nx, ny), periodic=periodic)
        )

    # Add connection in D3
    adjacency_matrix += np.eye(n_sites, k=nx * ny)

    # Wrap D3
    if periodic:
        adjacency_matrix += np.eye(n_sites, k=nx * ny * (nz - 1))

    adjacency_matrix += adjacency_matrix.T

    return np.array(adjacency_matrix, dtype=bool)


def hubbard_coefficients(
    n_modes: int,
    adjacency_matrix: npt.NDArray,
    onsite_term: float,
    hopping_term: float = 1.0,
    spinless: bool = False,
) -> tuple[np.ndarray, np.ndarray]:
    """Coefficients to fill a Hubbard Hamiltonian Template.

    Args:
        n_modes (int): Number of fermion modes in the system.
        adjacency_matrix (npt.NDArray): Adjacency matrix of lattice sites.
        onsite_term (float): Onsite interaction term.
        hopping_term (float): Kinetic term.
        spinless (bool): Set to True to use single spin Hamiltonian.

    Returns:
        tuple: one and two electron coefficients.
    """
    if not spinless:
        # We know which sites are adjacent, we need to restrict to same spin hopping.
        spin_adjacency_matrix = np.zeros(
            (2 * adjacency_matrix.shape[0], 2 * adjacency_matrix.shape[1])
        )
        spin_adjacency_matrix[::2, ::2] += adjacency_matrix
        spin_adjacency_matrix[1::2, 1::2] += adjacency_matrix
    else:
        spin_adjacency_matrix = adjacency_matrix

    one_e_coeffs = hopping_term * spin_adjacency_matrix
    one_e_coeffs = one_e_coeffs[:n_modes, :n_modes]

    two_e_coeffs = np.zeros((n_modes, n_modes, n_modes, n_modes))
    idx = np.arange(n_modes)
    two_e_coeffs[idx, idx, idx, idx] = onsite_term
    return one_e_coeffs, two_e_coeffs


def hubbard_hamiltonian(
    adjacency_matrix: npt.NDArray,
    onsite_term: float,
    hopping_term: float = 1.0,
    spinless: bool = False,
) -> FermionHamiltonian:
    """Return an encoded Hubbard hamiltonain with niave enumeration.

    As the Hubbard Hamiltonian has the same signature as the Chemists' Molecular Hamiltonian:
    (+-, +-+-)
    We can use the existing functions for the molecular Hamiltonian to create a template.

    Args:
        encoding (FermionQubitEncoding): The encoding to use.
        adjacency_matrix (npt.NDArray): Adjacency matrix of lattice sites.
        onsite_term (float): Onsite two-electron term.
        hopping_term (float): Kinetic term coefficient.
        physicist_noation (bool): Set to False for Chemist Notation.
        spinless (bool): Set to True to use single spin Hamiltonian.

    Returns:
        dict[str, float]: A qubit Hamiltonian.

    Example:
        >>> import numpy as np
        >>> from ferrmion.hamiltonians.molecular import hubbard_hamiltonian, square_adjacency_matrix
        >>> from ferrmion.encode import TernaryTree
        >>> tree = TernaryTree(12).JW()
        >>> adjacency = square_lattice_adjacency_matrix(shape=(4,4),periodic=False)
        >>> fham = hubbard_hamiltonian(adjacency, onsite_term=2,...)
        >>> tree.encode(fham)
    """
    n_sites = adjacency_matrix.shape[0]

    one_e_coeffs, two_e_coeffs = hubbard_coefficients(
        n_sites, adjacency_matrix, onsite_term, hopping_term, spinless=spinless
    )
    return FermionHamiltonian(
        terms={"+-": one_e_coeffs, "+-+-": two_e_coeffs}, constant_energy=0.0
    )



import numpy as np
import numpy.typing as npt

# Rust-accelerated functions exposed to Python

def symplectic_product(
    left: npt.NDArray[bool], right: npt.NDArray[bool]
) -> tuple[int, npt.NDArray[bool]]: ...
def hartree_fock_state(
    vacuum_state: npt.NDArray[np.float64],
    fermionic_hf_state: npt.NDArray[bool],
    mode_op_map: list[int],
    ipowers: npt.NDArray[np.uint8],
    symplectic_matrix: npt.NDArray[bool],
) -> tuple[npt.NDArray[np.complex128], npt.NDArray[bool]]: ...
def symplectic_to_pauli(symplectic: npt.NDArray[bool], int) -> tuple[str, int]: ...
def pauli_to_symplectic(pauli: str, ipower: int) -> tuple[npt.NDArray[bool], int]: ...
def symplectic_product_map(
    symplectics: npt.NDArray[bool],
    ipowers: npt.NDArray[np.uint8],
) -> tuple[npt.NDArray[np.uint8], npt.NDArray[bool]]: ...
def symplectic_to_sparse(
    symplectic: npt.NDArray[bool],
    ipower: int,
) -> tuple[str, npt.NDArray[np.uintp], np.complex]: ...
def molecular_hamiltonian_template(
    ipowers: npt.NDArray[np.uint8],
    symplectics: npt.NDArray[bool],
    physicist_notation: bool,
) -> dict: ...
def hubbard_hamiltonian_template(
    ipowers: npt.NDArray[np.uint8],
    symplectics: npt.NDArray[np.bool],
) -> dict: ...
def template_weight_distribution(
    template: dict,
    constant_energy: float,
    one_e_coeffs: npt.NDArray[np.float64],
    two_e_coeffs: npt.NDArray[np.float64],
    n_permutations: int,
) -> dict: ...
def fill_template(
    template: dict,
    constant_energy: float,
    one_e_coeffs: npt.NDArray[np.float64],
    two_e_coeffs: npt.NDArray[np.float64],
    mode_op_map: npt.NDArray[np.uint],
) -> dict: ...
def anneal_enumerations(
    ipowers: npt.NDArray[np.uint8],
    symplectics: npt.NDArray[np.bool],
    signatures: list[str],
    coeffs: list[np.ndarray],
    temperature: float,
    initial_guess: npt.NDArray[np.uint],
    coefficient_weighted: bool,
) -> tuple[float, npt.NDArray[np.uint]]: ...
def icount_to_sign(icount: int) -> np.complex64: ...
def topphatt(
    flatpack: list[tuple[np.uint, tuple[np.uint, np.uint, np.uint]]],
    n_qubits: int,
    signatures: list[str],
    coeffs: list[np.ndarray],
) -> tuple[npt.NDArray[np.uint8], npt.NDArray[np.bool]]: ...
def topphatt_standard(
    encoding: str,
    n_modes: int,
    n_qubits: int,
    signatures: list[str],
    coeffs: list[np.ndarray],
) -> tuple[npt.NDArray[np.uint8], npt.NDArray[np.bool]]: ...
def standard_symplectic_matrix(
    encoding: str, n_modes: int
) -> tuple[npt.NDArray[np.uint8], npt.NDArray[np.bool]]: ...
def encode_standard(
    encoding: str,
    n_modes: int,
    n_qubits: int,
    signatures: list[str],
    coeffs: list[np.ndarray],
    constant_energy:float,
) -> dict: ...
def encode(
    ipowers: npt.NDArray[np.uint8],
    symplectics: npt.NDArray[np.bool],
    signatures: list[str],
    coeffs: list[np.ndarray],
    constant_energy: float,
) -> dict: ...



"""Graph visualisation tools."""

import numpy as np
import rustworkx as rx
from rustworkx.visualization import mpl_draw

from ferrmion.encode import TernaryTree
from ferrmion.encode.ternary_tree_node import TTNode, node_sorter


def draw_tt(
    graph: rx.PyDiGraph | TTNode | TernaryTree,
    type: str = "spaced",
    enumeration_scheme=None,
    node_size=600,
    font_size=10,
):
    """Draws a rustworkx graph with nodes positioned as a ternary tree.

    Args:
        graph (rustworkx.PyDiGraph | ferrmion.TTNode | TernaryTree): A ternary tree.
        type (str): Make the graph prettier, one of "standard", "spaced", "linear".
        enumeration_scheme (dict[str, tuple[int, int]]): A mapping from node labels to a tuple of (mode index, qubit index).
        node_size (int): Size of nodes in the image.
        font_size (int): Size of node and edge labels in the image.

    Example:
        >>> from ferrmion.encode.ternary_tree import TernaryTree
        >>> from ferrmion.visualise.graph import draw_tt
        >>> tree = TernaryTree(3).Parity()
        >>> draw_tt(tree)
        >>> draw_tt(tree.root)
        >>> draw_tt(tree.root_node.to_rustworkx())
    """
    if isinstance(graph, TTNode):
        graph = graph.to_rustworkx()
    elif isinstance(graph, TernaryTree):
        graph = graph.root_node.to_rustworkx()

    def y_pos(label) -> float:
        return -3 * len(label)

    def x_pos(label) -> float:
        if type == "standard":
            pos = sum(
                [
                    (float(val) - 2) / (3**i)
                    for i, val in enumerate(list(str(node_sorter(label))))
                ]
            )
            pos = pos * len(label)

        elif type == "spaced":
            same_len = np.array([l for l in graph.nodes() if len(l) == len(label)])
            this_pos = np.where(same_len == label)[0][0]
            pos = (this_pos + 1) / (len(same_len) + 1) - 0.5

        elif type == "linear":
            same_len = np.array([l for l in graph.nodes() if len(l) == len(label)])
            this_pos = np.where(same_len == label)[0][0]
            pos = (this_pos + 1) / (len(same_len) + 1) - 0.5
            if len(same_len) <= 1:
                pos = len(label)
            if "z" not in label:
                pos *= -1.0
        else:
            raise ValueError("Type must be one of standard,spaced or linear.")
        return pos

    def format_label(label):
        return rf"$f_{{{enumeration_scheme[label][0]}}}q_{{{enumeration_scheme[label][1]}}}$"

    labels: callable = str if enumeration_scheme is None else format_label
    posmap = {
        index: [x_pos(label), y_pos(label)] for index, label in enumerate(graph.nodes())
    }
    posmap[0] = [0, 0]

    mpl_draw(
        graph,
        pos=posmap,
        with_labels=True,
        node_size=node_size,
        node_color="orange",
        edge_labels=str,
        labels=labels,
        font_size=font_size,
    )



"""Tests for Hamiltonan Functions."""
from ferrmion import TernaryTree
from ferrmion.hamiltonians import (
    molecular_hamiltonian,
    FermionHamiltonian,
    hubbard_hamiltonian
)
from ferrmion.core import encode_standard
import pytest
import numpy as np
from openfermion import QubitOperator, get_sparse_operator
from scipy.sparse.linalg import eigsh
from pytest import fixture
import logging
logger = logging.getLogger(__name__)


def test_molecular_hamiltonian_equivalent_explicit_fermion_hamiltonian():
    ones = np.eye(4)
    twos = np.ones((4,4,4,4))
    constant_energy = 10.
    molh = molecular_hamiltonian(one_e_coeffs=ones, two_e_coeffs=twos, constant_energy=constant_energy)
    explicit_molh = FermionHamiltonian()
    explicit_molh.creation().annihilation().with_coefficients(ones)
    explicit_molh.creation().creation().annihilation().annihilation().with_coefficients(twos)
    assert molh._terms.keys() == explicit_molh._terms.keys()
    assert np.all(molh._terms["+-"] == explicit_molh._terms["+-"])
    assert np.all(molh._terms["++--"] == explicit_molh._terms["++--"])

@pytest.mark.parametrize("encoding", ["JW", "BK", "PE", "JKMN"])
def test_encode_standard_water_eigvals_equal_expected(encoding, water_data):
    ones = water_data["ones"]
    twos = water_data["twos"]
    e_nuc = water_data["constant_energy"]

    qham = encode_standard(encoding, 14,14, ["+-","++--"], [ones, twos], e_nuc)

    ofop = QubitOperator()
    for k, v in qham.items():
        string = " ".join(
            [
                f"{char.upper()}{pos}" if char != "I" else ""
                for pos, char in enumerate(k)
            ]
        )
        ofop+= QubitOperator(term=string, coefficient=v)
    print(expected:=water_data["eigvals"])
    diag, _ = eigsh(get_sparse_operator(ofop), k=2, which="SA")
    print(diag)
    assert np.allclose(np.sort(diag), np.sort(expected)[:2])

@pytest.mark.parametrize("encoding", ["JW", "BK", "PE", "JKMN"])
def test_encode_standard_h2_eigvals_equal_expected(encoding, h2_mol_data_sets):
    ones = h2_mol_data_sets["ones"]
    twos = h2_mol_data_sets["twos"]
    e_nuc = h2_mol_data_sets["constant_energy"]
    n_modes = ones.shape[0]
    qham = encode_standard(encoding, n_modes, n_modes, ["+-","++--"], [ones, twos], e_nuc)

    ofop = QubitOperator()
    for k, v in qham.items():
        string = " ".join(
            [
                f"{char.upper()}{pos}" if char != "I" else ""
                for pos, char in enumerate(k)
            ]
        )
        ofop+= QubitOperator(term=string, coefficient=v)
    print(expected:=h2_mol_data_sets["eigvals"])
    diag, _ = eigsh(get_sparse_operator(ofop), k=2*n_modes, which="SA")
    print(diag)
    assert np.allclose(np.sort(diag), np.sort(h2_mol_data_sets["eigvals"]))



"""Tests for Simulated Annealing Optimisation."""
from typing import Callable
from ferrmion import TernaryTree

from ferrmion.optimize.topphatt import topphatt
from ferrmion.utils import fermionic_to_sparse_majorana
from ferrmion.encode.ternary_tree import (
    JordanWigner,
    BravyiKitaev,
    ParityEncoding,
    JKMN,
)
import numpy as np
import pytest
from ferrmion.core import anneal_enumerations, encode
from ferrmion.optimize.huffman import huffman_ternary_tree
from ferrmion.optimize.hatt import hamiltonian_adaptive_ternary_tree, fast_hatt
from openfermion import QubitOperator, get_sparse_operator
from scipy.sparse.linalg import eigsh



@pytest.mark.parametrize("encoding", [JordanWigner, ParityEncoding, BravyiKitaev, JKMN])
@pytest.mark.parametrize("coeff_weight", [True, False])
def test_core_topphatt_standard_h2_eigvals_equal_expected(encoding, coeff_weight, h2_mol_data_sets: dict):
    ones = h2_mol_data_sets["ones"]
    twos = h2_mol_data_sets["twos"]
    e_nuc = h2_mol_data_sets["constant_energy"]
    n_modes = ones.shape[0]

    ipow, sym = encoding(n_modes)._build_symplectic_matrix()
    anneal_enumerations(ipow, sym, ["+-","++--"], [ones, twos], n_modes, np.array([*range(n_modes)], dtype=np.uint), coeff_weight)
    qham = encode(ipow, sym, ["+-","++--"],[ones, twos], e_nuc)


    ofop = QubitOperator()
    for k, v in qham.items():
        string = " ".join(
            [
                f"{char.upper()}{pos}" if char != "I" else ""
                for pos, char in enumerate(k)
            ]
        )
        ofop+= QubitOperator(term=string, coefficient=v)
    print(expected:=h2_mol_data_sets["eigvals"])
    diag, _ = eigsh(get_sparse_operator(ofop), k=2*n_modes, which="SA")
    print(diag)
    assert np.allclose(np.sort(diag), np.sort(h2_mol_data_sets["eigvals"]))
